import pytest

from openff.interchange import Interchange
from openff.interchange.warnings import InterchangeDeprecationWarning


class TestDeprecation:
    @pytest.fixture
    def prepared_system(self, sage, water):
        return Interchange.from_smirnoff(sage, [water])

    def test_plus_operator_warning(self, monkeypatch, prepared_system):
        monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

        with pytest.warns(
            InterchangeDeprecationWarning,
            match="combine.*instead",
        ):
            prepared_system + prepared_system



"""Export to LAMMPS."""

from pathlib import Path
from typing import IO

import numpy
import packaging.version
from openff.toolkit.topology import Atom

from openff.interchange import Interchange
from openff.interchange.exceptions import UnsupportedExportError
from openff.interchange.interop.lammps.export.provenance import get_lammps_version
from openff.interchange.models import PotentialKey


def to_lammps(interchange: Interchange, file_path: Path | str, include_type_labels: bool = False):
    """Write an Interchange object to a LAMMPS data file."""
    if isinstance(file_path, str):
        path = Path(file_path)
    if isinstance(file_path, Path):
        path = file_path

    n_atoms = interchange.topology.n_atoms
    if "Bonds" in interchange.collections:
        n_bonds = len(interchange["Bonds"].key_map.keys())
    else:
        n_bonds = 0
    if "Angles" in interchange.collections:
        n_angles = len(interchange["Angles"].key_map.keys())
    else:
        n_angles = 0
    if "ProperTorsions" in interchange.collections:
        n_propers = len(interchange["ProperTorsions"].key_map.keys())
    else:
        n_propers = 0
    if "ImproperTorsions" in interchange.collections:
        n_impropers = len(interchange["ImproperTorsions"].key_map.keys())
    else:
        n_impropers = 0

    with open(path, "w") as lmp_file:
        lmp_file.write("Title\n\n")

        lmp_file.write(f"{n_atoms} atoms\n")
        lmp_file.write(f"{n_bonds} bonds\n")
        lmp_file.write(f"{n_angles} angles\n")
        lmp_file.write(f"{n_propers} dihedrals\n")
        lmp_file.write(f"{n_impropers} impropers\n")

        lmp_file.write(f"\n{len(interchange['vdW'].potentials)} atom types")
        if n_bonds > 0:
            lmp_file.write(f"\n{len(interchange['Bonds'].potentials)} bond types")
        if n_angles > 0:
            lmp_file.write(f"\n{len(interchange['Angles'].potentials)} angle types")
        if n_propers > 0:
            lmp_file.write(
                f"\n{len(interchange['ProperTorsions'].potentials)} dihedral types",
            )
        if n_impropers > 0:
            lmp_file.write(
                f"\n{len(interchange['ImproperTorsions'].potentials)} improper types",
            )

        lmp_file.write("\n")

        # write types section

        x_min, y_min, z_min = numpy.min(
            interchange.positions.to("angstrom"),  # type: ignore[union-attr]
            axis=0,
        ).magnitude
        if interchange.box is None:
            L_x, L_y, L_z = 100, 100, 100
        elif (interchange.box.m == numpy.diag(numpy.diagonal(interchange.box.m))).all():
            L_x, L_y, L_z = numpy.diag(interchange.box.to("angstrom").magnitude)
        else:
            raise NotImplementedError(
                "Interchange does not yet support exporting non-rectangular boxes to LAMMPS",
            )

        lmp_file.write(
            f"{x_min:.10g} {x_min + L_x:.10g} xlo xhi\n"
            f"{y_min:.10g} {y_min + L_y:.10g} ylo yhi\n"
            f"{z_min:.10g} {z_min + L_z:.10g} zlo zhi\n",
        )

        lmp_file.write("0.0 0.0 0.0 xy xz yz\n")

        if include_type_labels:
            # type labels added in 15Sep2022, see PR #1208
            assert get_lammps_version() > packaging.version.Version("2022.09.15")

            _write_atom_type_labels(lmp_file=lmp_file, interchange=interchange)
            if n_bonds > 0:
                _write_bond_type_labels(lmp_file=lmp_file, interchange=interchange)
            if n_angles > 0:
                _write_angle_type_labels(lmp_file=lmp_file, interchange=interchange)
            if n_propers > 0:
                _write_proper_type_labels(lmp_file=lmp_file, interchange=interchange)
            if n_impropers > 0:
                _write_improper_type_labels(lmp_file=lmp_file, interchange=interchange)

        lmp_file.write("\nMasses\n\n")

        vdw_handler = interchange["vdW"]
        atom_type_map = dict(enumerate(vdw_handler.potentials))
        key_map_inv = dict({v: k for k, v in vdw_handler.key_map.items()})

        for atom_type_idx, smirks in atom_type_map.items():
            # Find just one topology atom matching this SMIRKS by vdW
            matched_atom_idx = key_map_inv[smirks].atom_indices[0]
            matched_atom = interchange.topology.atom(matched_atom_idx)
            mass = matched_atom.mass.m

            lmp_file.write(f"{atom_type_idx + 1:d}\t{mass:.8g}\n")

        lmp_file.write("\n\n")

        _write_pair_coeffs(
            lmp_file=lmp_file,
            interchange=interchange,
            atom_type_map=atom_type_map,
        )

        if n_bonds > 0:
            _write_bond_coeffs(lmp_file=lmp_file, interchange=interchange)
        if n_angles > 0:
            _write_angle_coeffs(lmp_file=lmp_file, interchange=interchange)
        if n_propers > 0:
            _write_proper_coeffs(lmp_file=lmp_file, interchange=interchange)
        if n_impropers > 0:
            _write_improper_coeffs(lmp_file=lmp_file, interchange=interchange)

        _write_atoms(
            lmp_file=lmp_file,
            interchange=interchange,
            atom_type_map=atom_type_map,
        )
        if n_bonds > 0:
            _write_bonds(lmp_file=lmp_file, interchange=interchange)
        if n_angles > 0:
            _write_angles(lmp_file=lmp_file, interchange=interchange)
        if n_propers > 0:
            _write_propers(lmp_file=lmp_file, interchange=interchange)
        if n_impropers > 0:
            _write_impropers(lmp_file=lmp_file, interchange=interchange)


def _write_atom_type_labels(lmp_file: IO, interchange: Interchange):
    """Write the Atom Type Labels section of a LAMMPS data file."""
    lmp_file.write("\nAtom Type Labels\n\n")

    vdw_handler = interchange["vdW"]
    atom_type_map = dict(enumerate(vdw_handler.potentials))

    for atom_type_idx, smirks in atom_type_map.items():
        atom_type_label = f"{smirks.id}"
        lmp_file.write(f"{atom_type_idx + 1:d}\t{atom_type_label}\n")


def _write_bond_type_labels(lmp_file: IO, interchange: Interchange):
    """Write the Bond Type Labels section of a LAMMPS data file."""
    lmp_file.write("\nBond Type Labels\n\n")

    bond_handler = interchange["Bonds"]
    bond_type_map = dict(enumerate(bond_handler.potentials))

    for bond_type_idx, smirks in bond_type_map.items():
        bond_type_label = f"{smirks.id}"
        lmp_file.write(f"{bond_type_idx + 1:d}\t{bond_type_label}\n")


def _write_angle_type_labels(lmp_file: IO, interchange: Interchange):
    """Write the Angle Type Labels section of a LAMMPS data file."""
    lmp_file.write("\nAngle Type Labels\n\n")

    angle_handler = interchange["Angles"]
    angle_type_map = dict(enumerate(angle_handler.potentials))

    for angle_type_idx, smirks in angle_type_map.items():
        angle_type_label = f"{smirks.id}"
        lmp_file.write(f"{angle_type_idx + 1:d}\t{angle_type_label}\n")


def _write_proper_type_labels(lmp_file: IO, interchange: Interchange):
    """Write the Dihedral Type Labels section of a LAMMPS data file."""
    lmp_file.write("\nDihedral Type Labels\n\n")

    proper_handler = interchange["ProperTorsions"]
    proper_type_map = dict(enumerate(proper_handler.potentials))

    for proper_type_idx, smirks in proper_type_map.items():
        proper_type_label = f"{smirks.id}mult:{smirks.mult}"
        lmp_file.write(f"{proper_type_idx + 1:d}\t{proper_type_label}\n")


def _write_improper_type_labels(lmp_file: IO, interchange: Interchange):
    """Write the Improper Type Labels section of a LAMMPS data file."""
    lmp_file.write("\nImproper Type Labels\n\n")

    improper_handler = interchange["ImproperTorsions"]
    improper_type_map = dict(enumerate(improper_handler.potentials))

    for improper_type_idx, smirks in improper_type_map.items():
        improper_type_label = f"{smirks.id}"
        lmp_file.write(f"{improper_type_idx + 1:d}\t{improper_type_label}\n")


def _write_pair_coeffs(lmp_file: IO, interchange: Interchange, atom_type_map: dict):
    """Write the Pair Coeffs section of a LAMMPS data file."""
    lmp_file.write("Pair Coeffs\n\n")

    vdw_handler = interchange["vdW"]

    for atom_type_idx, smirks in atom_type_map.items():
        params = vdw_handler.potentials[smirks].parameters

        sigma = params["sigma"].to("angstrom").magnitude
        epsilon = params["epsilon"].to("kilocalorie / mole").magnitude

        lmp_file.write(f"{atom_type_idx + 1:d}\t{epsilon:.8g}\t{sigma:.8g}\n")

    lmp_file.write("\n")


def _write_bond_coeffs(lmp_file: IO, interchange: Interchange):
    """Write the Bond Coeffs section of a LAMMPS data file."""
    lmp_file.write("Bond Coeffs\n\n")

    bond_handler = interchange["Bonds"]
    bond_type_map = dict(enumerate(bond_handler.potentials))

    for bond_type_idx, smirks in bond_type_map.items():
        params = bond_handler.potentials[smirks].parameters

        k = params["k"].to("kilocalorie / mole / angstrom ** 2").magnitude
        k = k * 0.5  # Account for LAMMPS wrapping 1/2 into k
        length = params["length"].to("angstrom").magnitude

        lmp_file.write(f"{bond_type_idx + 1:d}\t{k:.16g}\t{length:.16g}\n")

    lmp_file.write("\n")


def _write_angle_coeffs(lmp_file: IO, interchange: Interchange):
    """Write the Angle Coeffs section of a LAMMPS data file."""
    lmp_file.write("\nAngle Coeffs\n\n")

    angle_handler = interchange["Angles"]
    angle_type_map = dict(enumerate(angle_handler.potentials))

    for angle_type_idx, smirks in angle_type_map.items():
        params = angle_handler.potentials[smirks].parameters

        k = params["k"].to("kilocalorie / mole / radian ** 2").magnitude
        k = k * 0.5  # Account for LAMMPS wrapping 1/2 into k
        theta = params["angle"].to("degree").magnitude

        lmp_file.write(f"{angle_type_idx + 1:d}\t{k:.16g}\t{theta:.16g}\n")

    lmp_file.write("\n")


def _write_proper_coeffs(lmp_file: IO, interchange: Interchange):
    """Write the Dihedral Coeffs section of a LAMMPS data file."""
    lmp_file.write("\nDihedral Coeffs\n\n")

    proper_handler = interchange["ProperTorsions"]
    proper_type_map = dict(enumerate(proper_handler.potentials))

    for proper_type_idx, smirks in proper_type_map.items():
        params = proper_handler.potentials[smirks].parameters

        k = params["k"].to("kilocalorie / mole").magnitude
        n = int(params["periodicity"])
        phase = params["phase"].to("degree").magnitude
        idivf = int(params["idivf"])
        k = k / idivf

        lmp_file.write(
            f"{proper_type_idx + 1:d} 1\t{k:.16g}\t{n:d}\t{phase:.16g}\n",
        )

    lmp_file.write("\n")


def _write_improper_coeffs(lmp_file: IO, interchange: Interchange):
    """Write the Improper Coeffs section of a LAMMPS data file."""
    lmp_file.write("\nImproper Coeffs\n\n")

    improper_handler = interchange["ImproperTorsions"]
    improper_type_map = dict(enumerate(improper_handler.potentials))

    for improper_type_idx, smirks in improper_type_map.items():
        params = improper_handler.potentials[smirks].parameters

        k = params["k"].to("kilocalorie / mole").magnitude
        n = int(params["periodicity"])
        phase = params["phase"].to("degree").magnitude
        idivf = int(params["idivf"])
        k = k / idivf

        # See https://lammps.sandia.gov/doc/improper_cvff.html
        # E_periodic = k * (1 + cos(n * theta - phase))
        # E_cvff = k * (1 + d * cos(n * theta))
        # k_periodic = k_cvff
        # if phase = 0,
        #   d_cvff = 1
        #   n_periodic = n_cvff
        # if phase = 180,
        #   cos(n * x - pi) == - cos(n * x)
        #   d_cvff = -1
        #   n_periodic = n_cvff
        # k * (1 + cos(n * phi - pi / 2)) == k * (1 - cos(n * phi))

        if phase == 0:
            k_cvff = k
            d_cvff = 1
            n_cvff = n
        elif phase == 180:
            k_cvff = k
            d_cvff = -1
            n_cvff = n
        else:
            raise UnsupportedExportError(
                "Improper exports to LAMMPS are funky and not well-supported, the only compatibility"
                "found between periodidic impropers is with improper_style cvff when phase = 0 or 180 degrees",
            )

        lmp_file.write(
            f"{improper_type_idx + 1:d} {k_cvff:.16g}\t{d_cvff:d}\t{n_cvff:.16g}\n",
        )

    lmp_file.write("\n")


def _write_atoms(lmp_file: IO, interchange: Interchange, atom_type_map: dict):
    """Write the Atoms section of a LAMMPS data file."""
    lmp_file.write("\nAtoms\n\n")

    atom_type_map_inv = dict({v: k for k, v in atom_type_map.items()})

    vdw_handler = interchange["vdW"]

    charges = interchange["Electrostatics"].charges
    positions = interchange.positions.m_as("angstrom")  # type: ignore[union-attr]

    """
    for molecule_index, molecule in enumerate(interchange.topology.molecules):
        for atom in molecule.atoms:
            atom._molecule_index = molecule_index
    """
    # Looking up an atom in a topology is slow, so establish a mapping between
    # atom and molecule indices by iterating through the topology once beforehand
    atom_map: dict[int, Atom] = dict()

    # the molecule is hashed for speed (bypassing an RDKit/OpenEye conversion)
    # didn't look into hashing atoms
    molecule_map: dict[int, int] = dict()
    atom_molecule_map: dict[Atom, int] = dict()

    for atom_index, atom in enumerate(interchange.topology.atoms):
        atom_map[atom_index] = atom

    for molecule_index, molecule in enumerate(interchange.topology.molecules):
        # inject mol ID into hash to allow chemically-identical Molecules
        # to be labelled with distinct IDs
        molecule_hash = hash(
            f"{molecule_index}_{molecule.ordered_connection_table_hash()}",
        )
        # TODO: see if this fix also applied for the same issue in GROMACS/AMBER writers

        molecule_map[molecule_hash] = molecule_index
        for atom in molecule.atoms:
            atom_molecule_map[atom] = molecule_hash

    def atom_index_to_molecule_index(atom_index: int) -> int:
        """Given an atom index (in the topology), return the molecule index."""
        return molecule_map[atom_molecule_map[atom_map[atom_index]]]

    for _, (top_key, pot_key) in enumerate(vdw_handler.key_map.items()):
        # TODO: Surely there's an easier way to get the molecule
        #       index from an atom index?
        atom_index = top_key.atom_indices[0]
        molecule_index = atom_index_to_molecule_index(atom_index)

        atom_type = atom_type_map_inv[pot_key]

        charge = charges[top_key].m

        pos = positions[atom_index]

        lmp_file.write(
            f"{atom_index + 1:d}\t"
            f"{molecule_index + 1:d}\t"
            f"{atom_type + 1:d}\t"
            f"{charge:.8g}\t"
            f"{pos[0]:.8g}\t"
            f"{pos[1]:.8g}\t"
            f"{pos[2]:.8g}\n",
        )


# Note: For this and the other valence writers, a significant portion of runtime
# appears to be spent looking up the LAMMPS "type" index from the potential key.
# (bond_map[pot_key]). Not sure why - I thought dictionary lookups were always
# fast - but it might be worth looking into.
def _write_bonds(lmp_file: IO, interchange: Interchange):
    """Write the Bonds section of a LAMMPS data file."""
    lmp_file.write("\nBonds\n\n")

    bond_handler = interchange["Bonds"]
    bond_map: dict[PotentialKey, int] = {
        potential_key: index for index, potential_key in enumerate(bond_handler.potentials)
    }

    for bond_index, (top_key, pot_key) in enumerate(bond_handler.key_map.items()):
        indices = top_key.atom_indices

        lmp_file.write(
            f"{bond_index + 1:d}\t{bond_map[pot_key] + 1:d}\t{indices[0] + 1:d}\t{indices[1] + 1:d}\n",
        )


def _write_angles(lmp_file: IO, interchange: Interchange):
    """Write the Angles section of a LAMMPS data file."""
    lmp_file.write("\nAngles\n\n")

    angle_handler = interchange["Angles"]
    angle_map: dict[PotentialKey, int] = {
        potential_key: index for index, potential_key in enumerate(angle_handler.potentials)
    }

    for angle_index, (top_key, pot_key) in enumerate(angle_handler.key_map.items()):
        indices = top_key.atom_indices

        lmp_file.write(
            f"{angle_index + 1:d}\t"
            f"{angle_map[pot_key] + 1:d}\t"
            f"{indices[0] + 1:d}\t"
            f"{indices[1] + 1:d}\t"
            f"{indices[2] + 1:d}\n",
        )


def _write_propers(lmp_file: IO, interchange: Interchange):
    """Write the Dihedrals section of a LAMMPS data file."""
    lmp_file.write("\nDihedrals\n\n")

    proper_handler = interchange["ProperTorsions"]
    proper_map: dict[PotentialKey, int] = {
        potential_key: index for index, potential_key in enumerate(proper_handler.potentials)
    }

    for proper_index, (top_key, pot_key) in enumerate(proper_handler.key_map.items()):
        indices = top_key.atom_indices

        lmp_file.write(
            f"{proper_index + 1:d}\t"
            f"{proper_map[pot_key] + 1:d}\t"
            f"{indices[0] + 1:d}\t"
            f"{indices[1] + 1:d}\t"
            f"{indices[2] + 1:d}\t"
            f"{indices[3] + 1:d}\n",
        )


def _write_impropers(lmp_file: IO, interchange: Interchange):
    """Write the Impropers section of a LAMMPS data file."""
    lmp_file.write("\nImpropers\n\n")

    improper_handler = interchange["ImproperTorsions"]
    improper_map: dict[PotentialKey, int] = {
        potential_key: index for index, potential_key in enumerate(improper_handler.potentials)
    }

    # Molecule/Topology.impropers lists the central atom SECOND,
    # but the improper collection lists the central atom FIRST,
    # However, at this point we're not looking in the topology directly,
    # we're assuming that the contents of the collection is encompassing
    for improper_index, (top_key, pot_key) in enumerate(
        improper_handler.key_map.items(),
    ):
        indices = top_key.atom_indices

        # https://github.com/openforcefield/openff-interchange/issues/544
        # LAMMPS, at least with `improper_style cvff`, lists the
        # central atom FIRST, which matches the collection
        lmp_file.write(
            f"{improper_index + 1:d}\t"
            f"{improper_map[pot_key] + 1:d}\t"
            f"{indices[0] + 1:d}\t"
            f"{indices[1] + 1:d}\t"
            f"{indices[2] + 1:d}\t"
            f"{indices[3] + 1:d}\n",
        )



import pathlib

import numpy
from openff.toolkit import Quantity

from openff.interchange._experimental import experimental
from openff.interchange.exceptions import GROMACSParseError
from openff.interchange.interop.gromacs._import.combine import make_monolithic
from openff.interchange.interop.gromacs.models.models import (
    GROMACSAngle,
    GROMACSAtom,
    GROMACSAtomType,
    GROMACSBond,
    GROMACSDihedral,
    GROMACSExclusion,
    GROMACSMolecule,
    GROMACSPair,
    GROMACSSettles,
    GROMACSSystem,
    LennardJonesAtomType,
    PeriodicImproperDihedral,
    PeriodicProperDihedral,
    RyckaertBellemansDihedral,
)


@experimental
def from_files(top_file, gro_file, cls=GROMACSSystem) -> GROMACSSystem:
    """
    Parse a GROMACS topology file. Adapted from Intermol.

    https://github.com/shirtsgroup/InterMol/blob/v0.1.2/intermol/gromacs/gromacs_parser.py
    """
    current_directive = "UNDEFINED"
    for line in make_monolithic(top_file):
        stripped = line.split(";")[0].strip()

        if len(stripped) == 0:
            continue

        if stripped.startswith(";"):
            continue

        if stripped.startswith("*"):
            continue

        if stripped.startswith("["):
            if not len(stripped.split()) == 3 and stripped.endswith("]"):
                raise GROMACSParseError("Invalid GROMACS topology file")

            current_directive = stripped[1:-1].strip()

            continue

        if current_directive == "defaults":
            (
                nonbonded_function,
                combination_rule,
                gen_pairs,
                vdw_14,
                coul_14,
            ) = _process_defaults(line)

            system = cls(
                nonbonded_function=nonbonded_function,
                combination_rule=combination_rule,
                gen_pairs=gen_pairs,
                vdw_14=vdw_14,
                coul_14=coul_14,
            )

        elif current_directive == "atomtypes":
            atom_type = _process_atomtype(line)
            system.atom_types[atom_type.name] = atom_type

        elif current_directive == "moleculetype":
            molecule_type = _process_moleculetype(line)
            system.molecule_types[molecule_type.name] = molecule_type

            current_molecule = molecule_type.name

        elif current_directive == "atoms":
            system.molecule_types[current_molecule].atoms.append(
                _process_atom(line),
            )

        elif current_directive == "pairs":
            system.molecule_types[current_molecule].pairs.append(
                _process_pair(line),
            )

        elif current_directive == "settles":
            system.molecule_types[current_molecule].settles.append(
                _process_settles(line),
            )

        elif current_directive == "bonds":
            system.molecule_types[current_molecule].bonds.append(
                _process_bond(line),
            )

        elif current_directive == "angles":
            system.molecule_types[current_molecule].angles.append(
                _process_angle(line),
            )

        elif current_directive == "dihedrals":
            system.molecule_types[current_molecule].dihedrals.append(
                _process_dihedral(line),
            )

        elif current_directive in ["exclusions"]:
            system.molecule_types[current_molecule].exclusions.append(
                _process_exclusion(line),
            )

        elif current_directive == "system":
            system.name = _process_system(line)

        elif current_directive == "molecules":
            molecule_name, number_of_copies = _process_molecule(line)

            system.molecules.append((molecule_name, number_of_copies))

        elif current_directive == "UNDEFINED":
            raise GROMACSParseError(
                "Failed to begin parsing GROMACS topology file. Error processing line:"
                f"\n\n{line}\n"
                "If you think this topology file is "
                "valid, please open an issue on GitHub.",
            )

        else:
            raise GROMACSParseError(f"Invalid directive {current_directive}")

    for molecule_type in system.molecule_types.values():
        this_molecule_atom_type_names = tuple(atom.atom_type for atom in molecule_type.atoms)

        molecule_type._contained_atom_types = {
            atom_type_name: system.atom_types[atom_type_name] for atom_type_name in this_molecule_atom_type_names
        }

    system.positions = _read_coordinates(gro_file)
    system.box = _read_box(gro_file)

    return system


def _process_defaults(line: str) -> tuple[int, int, str, float, float]:
    split = line.split()

    nonbonded_function = int(split[0])

    if nonbonded_function != 1:
        raise ValueError("Only LJ nonbonded functions are supported.")

    combination_rule = int(split[1])

    if combination_rule not in (2, 3):
        raise ValueError("combination rule 1 not supported.")

    gen_pairs = split[2]
    lj_14 = float(split[3])
    coul_14 = float(split[4])

    return nonbonded_function, combination_rule, gen_pairs, lj_14, coul_14


def _process_atomtype(
    line: str,
) -> GROMACSAtomType:
    split = line.split()

    # The second (bonding type) and third (atomic number) columns are optional. Handle these cases
    # explicitly by assuming that the ptype column is always a 1-length string.
    # See https://github.com/shirtsgroup/InterMol/blob/v0.1.2/intermol/gromacs/gromacs_parser.py#L1357-L1368

    if len(split[3]) == 1:
        # Bonded type and atomic number are both missing.
        split.insert(1, None)  # type: ignore[arg-type]
        split.insert(1, None)  # type: ignore[arg-type]

    elif len(split[4]) == 1 and len(split[5]) >= 1:
        if split[1][0].isalpha():
            # Atomic number is missing.
            split.insert(2, None)  # type: ignore[arg-type]
        else:
            # Bonded type is missing.
            split.insert(1, None)  # type: ignore[arg-type]

    atom_type = split[0]
    bonding_type = split[1] if split[1] else ""

    atomic_number = int(split[2]) if split[2] is not None else None
    mass = Quantity(float(split[3]), "dalton")

    charge = Quantity(float(split[4]), "elementary_charge")

    particle_type = split[5]

    if particle_type == "A":
        sigma = Quantity(float(split[6]), "nanometer")
        epsilon = Quantity(float(split[7]), "kilojoule_per_mole")
    else:
        raise ValueError(f"Particle type must be A, parsed {particle_type}.")

    return LennardJonesAtomType(
        name=atom_type,
        bonding_type=bonding_type,
        atomic_number=atomic_number,
        mass=mass,
        charge=charge,
        particle_type=particle_type,
        sigma=sigma,
        epsilon=epsilon,
    )


def _process_moleculetype(line: str) -> GROMACSMolecule:
    split = line.split()

    molecule_type = split[0]
    nrexcl = int(split[1])

    return GROMACSMolecule(name=molecule_type, nrexcl=nrexcl)


def _process_atom(
    line: str,
) -> GROMACSAtom:
    split = line.split()

    atom_number = int(split[0])
    atom_type = split[1]
    residue_number = int(split[2])
    residue_name = split[3]
    atom_name = split[4]
    charge_group_number = int(split[5])
    charge = Quantity(float(split[6]), "elementary_charge")
    mass = Quantity(float(split[7]), "amu")

    return GROMACSAtom(
        index=atom_number,
        atom_type=atom_type,
        name=atom_name,
        residue_index=residue_number,
        residue_name=residue_name,
        charge_group_number=charge_group_number,
        charge=charge,
        mass=mass,
    )


def _process_pair(line: str) -> GROMACSPair:
    split = line.split()

    atom1 = int(split[0])
    atom2 = int(split[1])

    nonbonded_function = int(split[2])

    if nonbonded_function == 1:
        return GROMACSPair(
            atom1=atom1,
            atom2=atom2,
        )
    else:
        raise ValueError(
            f"Nonbonded function must be 1, parsed a pair with {nonbonded_function}.",
        )


def _process_settles(line: str) -> GROMACSSettles:
    split = line.split()

    first_atom = int(split[0])

    oxygen_hydrogen_distance = Quantity(float(split[2]), "nanometer")
    hydrogen_hydrogen_distance = Quantity(float(split[3]), "nanometer")

    return GROMACSSettles(
        first_atom=first_atom,
        oxygen_hydrogen_distance=oxygen_hydrogen_distance,
        hydrogen_hydrogen_distance=hydrogen_hydrogen_distance,
    )


def _process_bond(line: str) -> GROMACSBond:
    split = line.split()

    atom1 = int(split[0])
    atom2 = int(split[1])

    bond_function = int(split[2])

    if bond_function == 1:
        bond_length = Quantity(float(split[3]), "nanometer")
        bond_k = Quantity(float(split[4]), "kilojoule_per_mole / nanometer**2")

        return GROMACSBond(
            atom1=atom1,
            atom2=atom2,
            function=bond_function,
            length=bond_length,
            k=bond_k,
        )

    else:
        raise ValueError(f"Bond function must be 1, parsed {bond_function}.")


def _process_angle(
    line: str,
) -> GROMACSAngle:
    split = line.split()

    atom1 = int(split[0])
    atom2 = int(split[1])
    atom3 = int(split[2])

    angle_function = int(split[3])

    if angle_function == 1:
        angle = Quantity(float(split[4]), "degrees")
        k = Quantity(float(split[5]), "kilojoule_per_mole")
    else:
        raise ValueError(f"Angle function must be 1, parsed {angle_function}.")

    return GROMACSAngle(
        atom1=atom1,
        atom2=atom2,
        atom3=atom3,
        angle=angle,
        k=k,
    )


def _process_dihedral(
    line: str,
) -> GROMACSDihedral:
    split = line.split()

    atom1 = int(split[0])
    atom2 = int(split[1])
    atom3 = int(split[2])
    atom4 = int(split[3])

    dihedral_function = int(split[4])

    if dihedral_function == 1:
        return PeriodicProperDihedral(
            atom1=atom1,
            atom2=atom2,
            atom3=atom3,
            atom4=atom4,
            phi=Quantity(float(split[5]), "degrees"),
            k=Quantity(float(split[6]), "kilojoule_per_mole"),
            multiplicity=int(float(split[7])),
        )

    elif dihedral_function == 3:
        return RyckaertBellemansDihedral(
            atom1=atom1,
            atom2=atom2,
            atom3=atom3,
            atom4=atom4,
            c0=Quantity(float(split[5]), "kilojoule_per_mole"),
            c1=Quantity(float(split[6]), "kilojoule_per_mole"),
            c2=Quantity(float(split[7]), "kilojoule_per_mole"),
            c3=Quantity(float(split[8]), "kilojoule_per_mole"),
            c4=Quantity(float(split[9]), "kilojoule_per_mole"),
            c5=Quantity(float(split[10]), "kilojoule_per_mole"),
        )

    elif dihedral_function == 4:
        return PeriodicImproperDihedral(
            atom1=atom1,
            atom2=atom2,
            atom3=atom3,
            atom4=atom4,
            phi=Quantity(float(split[5]), "degrees"),
            k=Quantity(float(split[6]), "kilojoule_per_mole"),
            multiplicity=int(float(split[7])),
        )

    else:
        raise ValueError(
            f"Dihedral functions 1, 3, and 4 supported, parsed {dihedral_function}.",
        )


def _process_exclusion(line: str) -> GROMACSExclusion:
    split = line.split()

    return GROMACSExclusion(
        first_atom=int(split[0]),
        other_atoms=[int(atom) for atom in split[1:]],
    )


def _process_molecule(line: str) -> tuple[str, int]:
    split = line.split()

    molecule_name = split[0]
    number_of_molecules = int(split[1])

    return molecule_name, number_of_molecules


def _process_system(line: str) -> str:
    split = line.split()

    system_name = split[0]

    return system_name


def _read_coordinates(file_path: pathlib.Path) -> Quantity:
    def _infer_coord_precision(file_path: pathlib.Path) -> int:
        """
        Infer decimal precision of coordinates by parsing periods in atoms lines.
        """
        with open(file_path) as file_in:
            file_in.readline()
            file_in.readline()
            atom_line = file_in.readline()
            period_indices = [i for i, x in enumerate(atom_line) if x == "."]
            spacing_between_periods = period_indices[-1] - period_indices[-2]
            precision = spacing_between_periods - 5
            return precision

    precision = _infer_coord_precision(file_path)
    coordinate_width = precision + 5
    # Column numbers in file separating x, y, z coords of each atom.
    # Default (3 decimals of precision -> 8 columns) are 20, 28, 36, 44
    coordinate_columns = [
        20,
        20 + coordinate_width,
        20 + 2 * coordinate_width,
        20 + 3 * coordinate_width,
    ]

    with open(file_path) as gro_file:
        # Throw away comment / name line
        gro_file.readline()
        n_atoms = int(gro_file.readline())

        unitless_coordinates = numpy.zeros((n_atoms, 3))
        for coordinate_index in range(n_atoms):
            line = gro_file.readline()
            _ = int(line[:5])  # residue_index
            _ = line[5:10]  # residue_name
            _ = line[10:15]  # atom_name
            _ = int(line[15:20])  # atom_index
            x = float(line[coordinate_columns[0] : coordinate_columns[1]])
            y = float(line[coordinate_columns[1] : coordinate_columns[2]])
            z = float(line[coordinate_columns[2] : coordinate_columns[3]])
            unitless_coordinates[coordinate_index] = numpy.array([x, y, z])

        coordinates = Quantity(unitless_coordinates, "nanometer")

    return coordinates


def _read_box(file_path: pathlib.Path) -> Quantity:
    with open(file_path) as gro_file:
        # Throw away comment / name line
        gro_file.readline()
        n_atoms = int(gro_file.readline())

        box_line = gro_file.readlines()[n_atoms]

    parsed_box = [float(val) for val in box_line.split()]

    if len(parsed_box) == 3:
        box = Quantity(parsed_box * numpy.eye(3), "nanometer")

    return box



"""
Test the behavior of the drivers.all module
"""

import math

import pandas
import pytest
from openff.toolkit import Quantity
from openff.utilities.testing import skip_if_missing

from openff.interchange._tests import (
    HAS_GROMACS,
    HAS_LAMMPS,
    HAS_SANDER,
    MoleculeWithConformer,
    needs_gmx,
    needs_lmp,
    needs_not_gmx,
    needs_not_lmp,
    needs_not_sander,
    needs_sander,
)
from openff.interchange.drivers.all import get_all_energies, get_summary_data


@skip_if_missing("openmm")
class TestDriversAll:
    @pytest.fixture
    def basic_interchange(self, sage_unconstrained):
        molecule = MoleculeWithConformer.from_smiles("CCO")
        molecule.name = "MOL"
        topology = molecule.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        return sage_unconstrained.create_interchange(topology)

    @needs_gmx
    @needs_lmp
    @needs_sander
    def test_all_with_all(self, basic_interchange):
        summary = get_all_energies(basic_interchange)

        assert len(summary) == 4

    @needs_not_gmx
    @needs_not_lmp
    @needs_not_sander
    def test_all_with_minimum(self, basic_interchange):
        summary = get_all_energies(basic_interchange)

        assert len(summary) == 1

    def test_skipping(self, basic_interchange):
        summary = get_all_energies(basic_interchange)

        assert ("GROMACS" in summary) == HAS_GROMACS
        assert ("Amber" in summary) == HAS_SANDER
        assert ("LAMMPS" in summary) == HAS_LAMMPS

    # TODO: Also run all of this with h-bond constraints
    def test_summary_data(self, basic_interchange):
        summary = get_summary_data(basic_interchange)

        if len(summary) < 2:
            pytest.skip("Not enough engines available to compare results")

        assert isinstance(summary, pandas.DataFrame)

        assert "OpenMM" in summary.index

        assert ("GROMACS" in summary.index) == HAS_GROMACS
        assert ("Amber" in summary.index) == HAS_SANDER
        assert ("LAMMPS" in summary.index) == HAS_LAMMPS

        # Check that (some of) the data is reasonable, this tolerance should be greatly reduced
        # See https://github.com/openforcefield/openff-interchange/issues/632
        for key in ["Bond", "Angle", "Torsion"]:
            assert summary.describe().loc["std", key] < 0.001, f"{key} failed comparison"

        # Check that (some of) the data did not NaN out
        for val in summary["Torsion"].to_dict().values():
            assert not math.isnan(val)



import pytest
from openff.toolkit import Quantity
from openff.units import ensure_quantity
from openff.utilities.testing import skip_if_missing

from openff.interchange.constants import kj_mol
from openff.interchange.drivers.report import EnergyReport
from openff.interchange.exceptions import (
    EnergyError,
    IncompatibleTolerancesError,
    InvalidEnergyError,
)


class TestEnergyReport:
    @pytest.fixture
    def report(self):
        return EnergyReport(
            energies={
                "Bond": 10 * kj_mol,
                "Angle": 10 * kj_mol,
                "Torsion": 2 * kj_mol,
                "vdW": 20 * kj_mol,
                "Electrostatics": -10 * kj_mol,
            },
        )

    @skip_if_missing("openmm")
    def test_coerce_to_quantity(self):
        assert isinstance(
            EnergyReport(
                energies={
                    "Bond": ensure_quantity(10 * kj_mol, "openmm"),
                },
            )["Bond"],
            Quantity,
        )

    def test_getitem(self, report):
        assert "Bond" in str(report)
        assert report["Bond"].units == kj_mol
        assert report["Nonbonded"] is None

        with pytest.raises(LookupError, match="type <class 'int'>"):
            report[0]

        assert report["Total"].m_as(kj_mol) == 32
        assert report.total_energy == report["Total"]

    def test_bad_constructor(self):
        with pytest.raises(InvalidEnergyError, match="foo not understood"):
            EnergyReport(
                energies={
                    "foo": 1 * kj_mol,
                },
            )

    def test_update(self, report):
        assert report["Bond"].m == 10.0
        report.update({"Bond": -200.0 * kj_mol})

        assert report["Bond"].m == -200.0

    @skip_if_missing("openmm")
    @pytest.mark.parametrize(
        "unit_system",
        ["openff", "openmm"],
    )
    def test_update_unit_types(self, report, unit_system):
        report.update(
            {
                "Bond": ensure_quantity(55.55 * kj_mol, unit_system),
            },
        )

        assert isinstance(report["Bond"], Quantity)
        assert report["Bond"].m_as(kj_mol) == pytest.approx(55.55)

    def test_bad_update(self, report):
        with pytest.raises(InvalidEnergyError, match="foo not understood"):
            report.update({"foo": 1 * kj_mol})

    def test_sub(self):
        a = EnergyReport(energies={"Torsion": 10 * kj_mol})
        b = EnergyReport(energies={"Torsion": 15 * kj_mol})
        c = EnergyReport(energies={"Torsion": 15 * kj_mol, "Nonbonded": 10 * kj_mol})

        assert (b - a)["Torsion"] == Quantity(5, "kilojoule / mol")

        with pytest.warns(UserWarning, match="Did not find key Nonbonded"):
            c - a

    def test_diff_identical(self, report):
        differences = report.diff(report)

        assert differences == {
            "Bond": 0 * kj_mol,
            "Angle": 0 * kj_mol,
            "Torsion": 0 * kj_mol,
            "vdW": 0 * kj_mol,
            "Electrostatics": 0 * kj_mol,
        }

    def test_diff_combined_nonbonded(self, report):
        single_nonbonded = EnergyReport(
            energies={
                "Bond": 10 * kj_mol,
                "Angle": 10 * kj_mol,
                "Torsion": 2 * kj_mol,
                "Nonbonded": 8 * kj_mol,
            },
        )

        differences = report.diff(single_nonbonded)

        assert "vdW" not in differences
        assert "Electrostatics" not in differences

        assert abs(differences["Nonbonded"].m) == 2

    def test_diff_multiple(self, report):
        other = EnergyReport(
            energies={
                "Bond": -4 * kj_mol,
                "Angle": 10 * kj_mol,
                "Torsion": -2 * kj_mol,
                "vdW": 10 * kj_mol,
                "Electrostatics": -10 * kj_mol,
            },
        )

        diff = report.diff(other)

        assert len(diff) == 5
        assert len([val for val in diff.values() if val.m != 0]) == 3

    def test_compare_identical(self, report):
        report.compare(report)

    def test_compare_different(self, report):
        other = EnergyReport(
            energies={
                key: val + 0.2 * kj_mol
                for key, val in {
                    "Bond": 10 * kj_mol,
                    "Angle": 10 * kj_mol,
                    "Torsion": 2 * kj_mol,
                    "vdW": 20 * kj_mol,
                    "Electrostatics": -10 * kj_mol,
                }.items()
            },
        )

        with pytest.raises(EnergyError):
            report.compare(other)

    def test_compare_multiple(self, report):
        other = EnergyReport(
            energies={
                "Bond": -4 * kj_mol,
                "Angle": 10 * kj_mol,
                "Torsion": -2 * kj_mol,
                "vdW": 10 * kj_mol,
                "Electrostatics": -10 * kj_mol,
            },
        )

        with pytest.raises(
            EnergyError,
            match=r"Bond.*Torsion.*vdW",
        ):
            report.compare(other)

    def test_compare_incompatible_tolerances(self, report):
        with pytest.raises(
            IncompatibleTolerancesError,
            match="whether nonbonded",
        ):
            report.compare(report, {"Nonbonded": 0.0 * kj_mol})



import os.path
import tempfile
import time

import polars
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from tqdm import tqdm

from openff.interchange import __version__

data_frame_file = f"{__version__}/mixed_solvent.csv"

if os.path.exists(data_frame_file):
    exit("CSV file already exists, exiting")

benzene = Molecule.from_smiles("c1ccccc1")
hexanol = Molecule.from_smiles("CCCCCCO")

benzene.generate_conformers(n_conformers=1)
hexanol.generate_conformers(n_conformers=1)

interchange = ForceField("openff-2.3.0-rc1.offxml").create_interchange(
    Topology.from_molecules(
        1000 * [benzene] + 100 * [hexanol],
    ),
)

interchange.box = Quantity([4, 4, 4], "nanometer")

times = list()

with tempfile.TemporaryDirectory() as tmpdir:
    for index in tqdm(range(10), desc="Mixed solvent benchmark"):
        openmm_start = time.perf_counter()
        interchange.to_openmm()
        openmm_end = time.perf_counter()

        gromacs_start = time.perf_counter()
        interchange.to_gromacs(prefix=f"{tmpdir}/temp")
        gromacs_end = time.perf_counter()

        times.append(
            {
                "iteration": index,
                "openmm_time": openmm_end - openmm_start,
                "gromacs_time": gromacs_end - gromacs_start,
            },
        )

data_frame = polars.from_dicts(
    data=times,
    schema={
        "iteration": polars.Int64,
        "openmm_time": polars.Float64,
        "gromacs_time": polars.Float64,
    },
)

data_frame.write_csv(data_frame_file)



"""Plugins handling virtual sites."""

from openff.interchange.components.potentials import Potential
from openff.interchange.components.toolkit import _validated_list_to_array
from openff.interchange.exceptions import InvalidParameterHandlerError
from openff.interchange.models import PotentialKey
from openff.interchange.smirnoff._nonbonded import SMIRNOFFElectrostaticsCollection
from openff.interchange.smirnoff._virtual_sites import SMIRNOFFVirtualSiteCollection
from openff.toolkit.typing.engines.smirnoff.parameters import (
    ParameterAttribute,
    VirtualSiteHandler,
    _BaseVirtualSiteType,
)
from openff.units import unit

from nonbonded_plugins.nonbonded import SMIRNOFFBuckinghamCollection


class BuckinghamVirtualSiteHandler(VirtualSiteHandler):
    """A handler for virtual sites compatible with the Buckingham (exp-6) functional form."""

    class BuckinghamVirtualSiteType(_BaseVirtualSiteType):
        """A type for virtual sites compatible with the Buckingham (exp-6) functional form."""

        _ELEMENT_NAME = "BuckinghamVirtualSite"

        _DEFAULT_A = 0.0 * unit.kilojoule_per_mole
        _DEFAULT_B = 0.0 * unit.nanometer**-1
        _DEFAULT_C = 0.0 * unit.kilojoule_per_mole * unit.nanometer**6

        # `unit` argument must be a Unit object, not a string
        a = ParameterAttribute(default=_DEFAULT_A, unit=_DEFAULT_A.units)
        b = ParameterAttribute(default=_DEFAULT_B, unit=_DEFAULT_B.units)
        c = ParameterAttribute(default=_DEFAULT_C, unit=_DEFAULT_C.units)

    _TAGNAME = "BuckinghamVirtualSites"
    _INFOTYPE = BuckinghamVirtualSiteType


class SMIRNOFFBuckinghamVirtualSiteCollection(SMIRNOFFVirtualSiteCollection):
    """A collection storing virtual sites compatible with the Buckingham (exp-6) functional form."""

    is_plugin: bool = True

    @classmethod
    def supported_parameters(cls):
        """Return a list of parameter attributes supported by this handler."""
        return [
            "type",
            "name",
            "id",
            "match",
            "smirks",
            "sigma",
            "epsilon",
            "rmin_half",
            "charge_increment",
            "distance",
            "outOfPlaneAngle",
            "inPlaneAngle",
        ]

    @classmethod
    def allowed_parameter_handlers(cls):
        """Return a list of allowed types of ParameterHandler classes."""
        return [BuckinghamVirtualSiteHandler]

    def store_potentials(  # type: ignore[override]
        self,
        parameter_handler: VirtualSiteHandler,
        vdw_collection: SMIRNOFFBuckinghamCollection,
        electrostatics_collection: SMIRNOFFElectrostaticsCollection,
    ) -> None:
        """Store VirtualSite-specific parameter-like data."""
        if self.potentials:
            self.potentials = dict()
        for virtual_site_key, potential_key in self.key_map.items():
            # TODO: This logic assumes no spaces in the SMIRKS pattern, name or `match` attribute
            smirks, _, _ = potential_key.id.split(" ")
            parameter = parameter_handler.parameters[smirks]

            virtual_site_potential = Potential(
                parameters={
                    "distance": parameter.distance,
                },
            )
            for attr in ["outOfPlaneAngle", "inPlaneAngle"]:
                if hasattr(parameter, attr):
                    virtual_site_potential.parameters.update(
                        {attr: getattr(parameter, attr)},
                    )
            self.potentials[potential_key] = virtual_site_potential

            vdw_key = PotentialKey(id=potential_key.id, associated_handler="vdw")
            vdw_potential = Potential(
                parameters={
                    parameter_name: getattr(parameter, parameter_name) for parameter_name in self.specific_parameters()
                },
            )
            vdw_collection.key_map[virtual_site_key] = vdw_key
            vdw_collection.potentials[vdw_key] = vdw_potential

            electrostatics_key = PotentialKey(
                id=potential_key.id,
                associated_handler="Electrostatics",
            )
            electrostatics_potential = Potential(
                parameters={
                    "charge_increments": _validated_list_to_array(
                        parameter.charge_increment,
                    ),
                },
            )
            electrostatics_collection.key_map[virtual_site_key] = electrostatics_key
            electrostatics_collection.potentials[electrostatics_key] = electrostatics_potential

    @classmethod
    def create(
        cls,
        parameter_handler,
        topology,
        vdw_collection,
        electrostatics_collection,
    ):
        """
        Create a SMIRNOFFCOllection from toolkit data.
        """
        if type(parameter_handler) not in cls.allowed_parameter_handlers():
            raise InvalidParameterHandlerError(type(parameter_handler))

        collection = cls()

        if hasattr(collection, "fractional_bondorder_method"):
            raise NotImplementedError(
                "Plugins with fractional bond order not yet supported",
            )

        collection.store_matches(parameter_handler=parameter_handler, topology=topology)
        collection.store_potentials(
            parameter_handler=parameter_handler,
            vdw_collection=vdw_collection,
            electrostatics_collection=electrostatics_collection,
        )

        return collection

    @classmethod
    def specific_parameters(cls) -> list[str]:
        """Parameters specific to this collection."""
        return ["a", "b", "c"]



import pytest
from openff.packmol import solvate_topology
from openff.toolkit import Quantity, Topology

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer
from openff.interchange.constants import kj_mol
from openff.interchange.drivers.openmm import _get_openmm_energies, _process, get_openmm_energies


class TestTIP4PVirtualSites:
    def test_tip4p_openmm_xml(self, water_dimer):
        """
        Prepare a TIP4P water dimer with OpenMM's style of 4-site water.

        Below is used as a guide
        https://openmm.github.io/openmm-cookbook/latest/notebooks/tutorials/Histone_methyltransferase_simulation_with_a_multisite_water_model_TIP4P-Ew.html
        """
        pytest.importorskip("openmm")

        import openmm.app

        modeller = openmm.app.Modeller(
            topology=water_dimer.to_openmm(),
            positions=water_dimer.get_positions().to("nanometer").to_openmm(),
        )

        forcefield = openmm.app.ForceField("tip4pew.xml")

        modeller.addExtraParticles(forcefield=forcefield)

        system = forcefield.createSystem(
            modeller.topology,
            nonbondedMethod=openmm.app.PME,
            nonbondedCutoff=1.0 * openmm.unit.nanometers,
            constraints=openmm.app.HBonds,
            rigidWater=True,
            ewaldErrorTolerance=0.0001,  # match Interchange.to_openmm_system default
        )

        imported = Interchange.from_openmm(
            topology=modeller.topology,
            system=system,
            positions=modeller.getPositions(),
            box_vectors=modeller.getTopology().getPeriodicBoxVectors(),
        )

        get_openmm_energies(
            imported,
            combine_nonbonded_forces=True,
        ).compare(
            _process(
                raw_energies=_get_openmm_energies(
                    system=system,
                    positions=modeller.getPositions(),
                    box_vectors=modeller.getTopology().getPeriodicBoxVectors(),
                    round_positions=None,
                    platform="Reference",
                ),
                combine_nonbonded_forces=True,
                detailed=False,
                system=system,
            ),
            tolerances={
                "Nonbonded": 1e-5 * kj_mol,
            },
        )

    @pytest.mark.skip(
        reason="Rewrite to use OpenMM or update from_openmm to support `LocalCoordinatesSite`s",
    )
    def test_dimer_energy_equals(self, tip4p, water_dimer):
        out: Interchange = tip4p.create_interchange(water_dimer)

        roundtripped = Interchange.from_openmm(
            system=out.to_openmm_system(),
            topology=out.to_openmm_topology(collate=False),
            positions=out.get_positions(include_virtual_sites=True).to_openmm(),
            box_vectors=out.box.to_openmm(),
        )

        assert get_openmm_energies(out) == _get_openmm_energies(roundtripped)

    @pytest.mark.skip(
        reason="Rewrite to use OpenMM or update from_openmm to support `LocalCoordinatesSite`s",
    )
    def test_minimize_solvated_ligand(self, sage_with_tip4p, default_integrator):
        topology = solvate_topology(
            topology=MoleculeWithConformer.from_smiles("CO").to_topology(),
            nacl_conc=Quantity(1.0, "mole / liter"),
            target_density=Quantity(0.6, "gram / milliliter"),
            working_directory=".",
        )

        simulation = sage_with_tip4p.create_interchange(topology).to_openmm_simulation(integrator=default_integrator)

        roundtripped = Interchange.from_openmm(
            system=simulation.system,
            topology=simulation.topology,
            positions=simulation.context.getState(getPositions=True).getPositions(),
            box_vectors=simulation.system.getDefaultPeriodicBoxVectors(),
        )

        original_energy = get_openmm_energies(simulation)

        # TODO: Much more validation could be done here, but if a simulation
        #       can start and minimize at all, that should catch most problems
        roundtripped.minimize()

        assert get_openmm_energies(roundtripped) < original_energy

    @pytest.mark.skip(
        reason=(
            "Doesn't work cleanly until importing LocalCoordinatesSite is supported, even though "
            "this error is probably raised before that type check"
        ),
    )
    def test_error_index_mismatch(self, tip4p, water):
        out: Interchange = tip4p.create_interchange(Topology.from_molecules([water, water]))

        with pytest.raises(
            ValueError,  # TODO: Make a different error
            match="Particle ordering mismatch",
        ):
            Interchange.from_openmm(
                system=out.to_openmm_system(),
                topology=out.to_openmm_topology(collate=True),
            )



import pytest
from openff.toolkit import ForceField, Molecule


def test_impropers_can_exist_with_no_propers(sage):
    """See issue #1441"""
    openmm = pytest.importorskip("openmm")

    ff = ForceField()
    ff.register_parameter_handler(sage.get_parameter_handler("ImproperTorsions"))

    topology = Molecule.from_smiles("c1ccccc1").to_topology()
    assert len(ff.label_molecules(topology)[0]["ImproperTorsions"])

    system = ff.create_interchange(topology).to_openmm(0)
    assert system.getNumParticles() == 12

    force_names = set(
        [system.getForce(i).__class__.__name__ for i in range(system.getNumForces())],
    )

    assert system.getNumForces() == 2, f"Expected 2 forces, got {system.getNumForces()}: {force_names}"

    assert force_names == {"CMMotionRemover", "PeriodicTorsionForce"}

    for force in system.getForces():
        if type(force) is not openmm.PeriodicTorsionForce:
            continue

        # probably don't need to inspect parameters themselves, just there
        # being a force and it having the right number of torsions is probably enough
        assert force.getNumTorsions() == 18



import numpy
from openff.toolkit import Molecule, Quantity, Topology
from openff.utilities.testing import skip_if_missing

from openff.interchange.components.mbuild import offmol_to_compound, offtop_to_compound


@skip_if_missing("mbuild")
class TestMBuildConversions:
    def test_basic_mol_to_compound(self):
        """Test basic behavior of conversion to mBuild Compound"""
        offmol = Molecule.from_smiles("CCO")
        offmol.generate_conformers(n_conformers=1)

        comp = offmol_to_compound(offmol)

        assert comp.n_particles == offmol.n_atoms
        assert comp.n_bonds == offmol.n_bonds

        numpy.testing.assert_equal(
            offmol.conformers[0].m_as("nanometer"),
            comp.xyz,
        )

    def test_mbuild_conversion_generate_conformers(self):
        """Test that a single conformer is automatically generated"""
        offmol = Molecule.from_smiles("CCO")

        comp = offmol_to_compound(offmol)

        assert comp.n_particles == offmol.n_atoms
        assert comp.n_bonds == offmol.n_bonds

        offmol.generate_conformers(n_conformers=1)
        expected_conf = offmol.conformers[0]

        numpy.testing.assert_equal(
            expected_conf.m_as("nanometer"),
            comp.xyz,
        )

    def test_mbuild_conversion_first_conformer_used(self):
        """Test that only the first conformer in an OFFMol is used"""
        offmol = Molecule.from_smiles("C1=CC=C(C=C1)C2=CC=C(C=C2)C3=CC=CC=C3")
        offmol.generate_conformers(n_conformers=3, rms_cutoff=Quantity(0.0, "angstrom"))

        comp = offmol_to_compound(offmol)

        numpy.testing.assert_equal(
            offmol.conformers[0].m_as("nanometer"),
            comp.xyz,
        )

        with numpy.testing.assert_raises(AssertionError):
            numpy.testing.assert_equal(
                offmol.conformers[1].m_as("nanometer"),
                comp.xyz,
            )

        with numpy.testing.assert_raises(AssertionError):
            numpy.testing.assert_equal(
                offmol.conformers[2].m_as("nanometer"),
                comp.xyz,
            )

    def test_mbuild_conversion_element_names(self):
        """Test that the generated Compound has particle names that can be
        interpreted as elements"""
        offmol = Molecule.from_smiles(
            "CSC(CO[P]([O-])([O-])=O)c1cc(Br)c(F)c(Cl)n1",
            allow_undefined_stereo=True,
        )
        comp = offmol_to_compound(offmol)

        known_elements = {"C", "H", "O", "N", "Cl", "Br", "F", "S", "P"}

        for particle in comp.particles():
            assert particle.name in known_elements

    def test_multi_mol_topology_to_compound(self):
        """Test the basic behavior of a (multi-mol) OFFTop"""
        ethanol = Molecule.from_smiles("CCO")
        ethanol.name = "ETH"
        methane = Molecule.from_smiles("C")
        methane.name = "MET"

        top = Topology.from_molecules([ethanol, ethanol, methane, methane])

        comp = offtop_to_compound(top)

        assert comp.n_particles == 28  # 2 * (9 + 5)
        assert comp.n_bonds == 24  # 2 * (8 + 4)
        assert len(comp.children) == 4  # 4 "molecules"



"""Models for non-standard non-bonded treatments."""

from typing import Literal

from openff.interchange.components.potentials import Collection


class BuckinghamvdWCollection(Collection):
    """Handler storing Buckingham-style vdW potentials."""

    type: Literal["Buckingham-6"] = "Buckingham-6"
    expression: Literal["a*exp(-b*r)-c*r**-6"] = "a*exp(-b*r)-c*r**-6"
    mixing_rule: Literal["buckingham"] = "buckingham"
    method: str = "cutoff"
    cutoff: float = 9.0
    scale_12: float = 0.0
    scale_13: float = 0.0
    scale_14: float = 0.5
    scale_15: float = 1.0



"""Script to generate a subset of MiniDrugBank."""

from openff.toolkit.topology import Molecule
from openff.toolkit.utils import get_data_file_path
from rdkit import Chem

trimmed_mol_names = [
    "DrugBank_5354",
    "DrugBank_2800",
    "DrugBank_5418",
    "DrugBank_104",
    "DrugBank_5516",
    "DrugBank_5523",
    "DrugBank_2967",
    "DrugBank_246",
    "DrugBank_2991",
    "DrugBank_3028",
    "DrugBank_3046",
    "DrugBank_3087",
    "DrugBank_390",
    "DrugBank_5804",
    "DrugBank_3346",
    "DrugBank_3358",
    "DrugBank_3406",
    "DrugBank_5900",
    "DrugBank_5902",
    "DrugBank_3479",
    "DrugBank_3503",
    "DrugBank_3547",
    "DrugBank_3565",
    "DrugBank_6032",
    "DrugBank_914",
    "DrugBank_977",
    "DrugBank_6182",
    "DrugBank_3817",
    "DrugBank_3954",
    "DrugBank_6355",
    "DrugBank_4074",
    "DrugBank_1448",
    "DrugBank_1449",
    "DrugBank_4138",
    "DrugBank_4161",
    "DrugBank_1538",
    "DrugBank_6531",
    "DrugBank_1564",
    "DrugBank_6533",
    "DrugBank_4215",
    "DrugBank_4217",
    "DrugBank_1598",
    "DrugBank_1608",
    "DrugBank_4249",
    "DrugBank_1637",
    "DrugBank_1661",
    "DrugBank_6647",
    "DrugBank_4323",
    "DrugBank_1721",
    "DrugBank_1722",
    "DrugBank_1742",
    "DrugBank_6722",
    "DrugBank_4468",
    "DrugBank_4515",
    "DrugBank_6865",
    "DrugBank_4580",
    "DrugBank_1971",
    "DrugBank_4662",
    "DrugBank_7108",
    "DrugBank_7124",
    "DrugBank_2140",
    "DrugBank_2148",
    "DrugBank_2186",
    "DrugBank_2237",
    "DrugBank_4959",
    "DrugBank_2429",
    "DrugBank_5154",
    "DrugBank_2563",
    "DrugBank_2570",
    "DrugBank_2584",
    "DrugBank_2585",
    "DrugBank_2684",
]


mini_drug_bank = Molecule.from_file(
    get_data_file_path("molecules/MiniDrugBank.sdf"),
    allow_undefined_stereo=True,
)
mols = [mol for mol in mini_drug_bank if mol.name in trimmed_mol_names]

with Chem.SDWriter("MiniDrugBankTrimmed.sdf") as f:
    for openff_mol in mols:
        try:
            openff_mol.assign_partial_charges(partial_charge_method="am1bcc")
            rdmol = openff_mol.to_rdkit()
        except:  # noqa
            continue
        f.write(rdmol)



import logging
import math
from typing import Literal

import numpy
from openff.toolkit import Quantity, Topology
from openff.toolkit.typing.engines.smirnoff.parameters import (
    ParameterHandler,
    VirtualSiteHandler,
)
from pydantic import Field

from openff.interchange._annotations import _DegreeQuantity, _Quantity
from openff.interchange.components._particles import _VirtualSite
from openff.interchange.components.potentials import Potential
from openff.interchange.components.toolkit import (
    _lookup_virtual_site_parameter,
    _validated_list_to_array,
)
from openff.interchange.interop._virtual_sites import _ThreeParticleAverageSite
from openff.interchange.models import BaseVirtualSiteKey, ImportedVirtualSiteKey, PotentialKey, SMIRNOFFVirtualSiteKey
from openff.interchange.smirnoff._base import SMIRNOFFCollection
from openff.interchange.smirnoff._nonbonded import (
    SMIRNOFFElectrostaticsCollection,
    SMIRNOFFvdWCollection,
)

logger = logging.getLogger(__name__)

_DEGREES_TO_RADIANS = numpy.pi / 180.0


# The use of `type` as a field name conflicts with the built-in `type()` when used with PEP 585
_ListOfHandlerTypes = list[type[ParameterHandler]]


class SMIRNOFFVirtualSiteCollection(SMIRNOFFCollection):
    """
    A handler which stores the information necessary to construct virtual sites (virtual particles).
    """

    key_map: dict[BaseVirtualSiteKey, PotentialKey] = Field(
        dict(),
        description="A mapping between VirtualSiteKey objects and PotentialKey objects.",
    )  # type: ignore[assignment]

    type: Literal["VirtualSites"] = "VirtualSites"
    expression: Literal[""] = ""
    virtual_site_key_topology_index_map: dict[BaseVirtualSiteKey, int] = Field(
        dict(),
        description="A mapping between VirtualSiteKey objects (stored analogously to TopologyKey objects"
        "in other handlers) and topology indices describing the associated virtual site",
    )
    exclusion_policy: Literal[
        "none",
        "minimal",
        "parents",
        "local",
        "neighbors",
        "connected",
        "all",
    ] = "parents"

    @classmethod
    def allowed_parameter_handlers(cls) -> _ListOfHandlerTypes:
        """Return a list of allowed types of ParameterHandler classes."""
        return [VirtualSiteHandler]

    @classmethod
    def supported_parameters(cls) -> list[str]:
        """Return a list of parameter attributes supported by this handler."""
        return [
            "type",
            "name",
            "id",
            "match",
            "smirks",
            "sigma",
            "epsilon",
            "rmin_half",
            "charge_increment",
            "distance",
            "outOfPlaneAngle",
            "inPlaneAngle",
        ]

    @classmethod
    def nonbonded_parameters(cls) -> list[str]:
        """Return a list of parameter attributes handling vdW interactions."""
        return ["sigma", "epsilon"]

    def store_matches(
        self,
        parameter_handler: ParameterHandler,
        topology: Topology,
    ) -> None:
        """Populate self.key_map with key-val pairs of [VirtualSiteKey, PotentialKey]."""
        if self.key_map:
            self.key_map = dict()

        # Initialze the virtual site index to begin after the topoogy's atoms (0-indexed)
        virtual_site_index = topology.n_atoms

        matches_by_parent = parameter_handler._find_matches_by_parent(topology)

        for parent_index, parameters in matches_by_parent.items():
            for parameter, orientations in parameters:
                for orientation in orientations:
                    orientation_indices = orientation.topology_atom_indices

                    virtual_site_key = SMIRNOFFVirtualSiteKey(
                        parent_atom_index=parent_index,
                        orientation_atom_indices=orientation_indices,
                        type=parameter.type,
                        name=parameter.name,
                        match=parameter.match,
                    )

                    # TODO: Better way of specifying unique parameters
                    potential_key = PotentialKey(
                        id=" ".join(
                            [parameter.smirks, parameter.name, parameter.match],
                        ),
                        associated_handler="VirtualSites",
                        virtual_site_type=parameter.type,
                    )
                    self.key_map[virtual_site_key] = potential_key
                    self.virtual_site_key_topology_index_map[virtual_site_key] = virtual_site_index
                    virtual_site_index += 1

    def store_potentials(  # type: ignore[override]
        self,
        parameter_handler: VirtualSiteHandler,
        vdw_collection: SMIRNOFFvdWCollection,
        electrostatics_collection: SMIRNOFFElectrostaticsCollection,
    ):
        """
        Populate self.potentials with key-val pairs of [PotentialKey, Potential].
        """
        if self.potentials:
            self.potentials = dict()
        for virtual_site_key, potential_key in self.key_map.items():
            # TODO: This logic assumes no spaces in the SMIRKS pattern, name or `match` attribute
            smirks, name, match = potential_key.id.split(" ")
            parameter = _lookup_virtual_site_parameter(
                parameter_handler=parameter_handler,
                smirks=smirks,
                name=name,
                match=match,
            )

            virtual_site_potential = Potential(
                parameters={
                    "distance": parameter.distance,
                },
            )
            for attr in ["outOfPlaneAngle", "inPlaneAngle"]:
                if hasattr(parameter, attr):
                    virtual_site_potential.parameters.update(
                        {attr: getattr(parameter, attr)},
                    )
            self.potentials[potential_key] = virtual_site_potential

            vdw_key = PotentialKey(id=potential_key.id, associated_handler="vdw")
            vdw_potential = Potential(
                parameters={
                    parameter_name: getattr(parameter, parameter_name)
                    for parameter_name in self.nonbonded_parameters()
                },
            )
            vdw_collection.key_map[virtual_site_key] = vdw_key
            vdw_collection.potentials[vdw_key] = vdw_potential

            electrostatics_key = PotentialKey(
                id=potential_key.id,
                associated_handler="Electrostatics",
            )
            electrostatics_potential = Potential(
                parameters={
                    "charge_increments": _validated_list_to_array(
                        parameter.charge_increment,
                    ),
                },
            )

            logger.debug(
                "Charge section VirtualSites applied to virtual site with orientation atoms at topology indices "
                f"{virtual_site_key.orientation_atom_indices}",
            )

            electrostatics_collection.key_map[virtual_site_key] = electrostatics_key
            electrostatics_collection.potentials[electrostatics_key] = electrostatics_potential


class _BondChargeVirtualSite(_VirtualSite):
    type: Literal["BondCharge"]
    distance: _Quantity
    orientations: tuple[int, ...]

    @property
    def local_frame_weights(self) -> tuple[list[float], ...]:
        origin_weight = [1.0, 0.0]  # first atom is origin
        x_direction = [-1.0, 1.0]
        y_direction = [-1.0, 1.0]

        return origin_weight, x_direction, y_direction

    @property
    def local_frame_positions(self) -> Quantity:
        distance_unit = self.distance.units
        return Quantity(
            [-self.distance.m, 0.0, 0.0],
            distance_unit,
        )

    @property
    def local_frame_coordinates(self) -> Quantity:
        return Quantity(
            numpy.array(
                [self.distance.m_as("nanometer"), 180.0, 0.0],
            ),
        )


class _MonovalentLonePairVirtualSite(_VirtualSite):
    type: Literal["MonovalentLonePair"]
    distance: _Quantity
    out_of_plane_angle: _DegreeQuantity
    in_plane_angle: _DegreeQuantity
    orientations: tuple[int, ...]

    @property
    def local_frame_weights(self) -> tuple[list[float], ...]:
        origin_weight = [1.0, 0.0, 0.0]  # first/zeroth atom is origin
        x_direction = [-1.0, 1.0, 0.0]
        y_direction = [-1.0, 0.0, 1.0]

        return origin_weight, x_direction, y_direction

    @property
    def local_frame_positions(self) -> Quantity:
        theta = self.in_plane_angle.m_as("radian")
        phi = self.out_of_plane_angle.m_as("radian")

        distance_unit = self.distance.units

        return Quantity(
            [
                self.distance.m * math.cos(theta) * math.cos(phi),
                self.distance.m * math.sin(theta) * math.cos(phi),
                self.distance.m * math.sin(phi),
            ],
            distance_unit,
        )

    @property
    def local_frame_coordinates(self) -> Quantity:
        return Quantity(
            numpy.array(
                [
                    self.distance.m_as("nanometer"),
                    self.in_plane_angle.m_as("degree"),
                    self.out_of_plane_angle.m_as("degree"),
                ],
            ),
        )


class _DivalentLonePairVirtualSite(_VirtualSite):
    type: Literal["DivalentLonePair"]
    distance: _Quantity
    out_of_plane_angle: _DegreeQuantity
    orientations: tuple[int, ...]

    @property
    def local_frame_weights(self) -> tuple[list[float], ...]:
        origin_weight = [1.0, 0.0, 0.0]  # first atom is origin
        x_direction = [-1.0, 0.5, 0.5]
        y_direction = [-1.0, 1.0, 0.0]

        return origin_weight, x_direction, y_direction

    @property
    def local_frame_positions(self) -> Quantity:
        theta = self.out_of_plane_angle.m_as("radian")

        distance_unit = self.distance.units

        return Quantity(
            [
                -self.distance.m * math.cos(theta),
                0.0,
                self.distance.m * math.sin(theta),
            ],
            distance_unit,
        )

    @property
    def local_frame_coordinates(self) -> Quantity:
        return Quantity(
            numpy.array(
                [
                    self.distance.m_as("nanometer"),
                    180.0,
                    self.out_of_plane_angle.m_as("degree"),
                ],
            ),
        )


class _TrivalentLonePairVirtualSite(_VirtualSite):
    type: Literal["TrivalentLonePair"]
    distance: _Quantity
    orientations: tuple[int, ...]

    @property
    def local_frame_weights(self) -> tuple[list[float], ...]:
        origin_weight = [1.0, 0.0, 0.0, 0.0]  # first atom is origin
        x_direction = [-1.0, 1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0]
        y_direction = [-1.0, 1.0, 0.0, 0.0]  # Not used (?)

        return origin_weight, x_direction, y_direction

    @property
    def local_frame_positions(self) -> Quantity:
        distance_unit = self.distance.units
        return Quantity(
            [-self.distance.m, 0.0, 0.0],
            distance_unit,
        )

    @property
    def local_frame_coordinates(self) -> Quantity:
        return Quantity(
            numpy.array(
                [self.distance.m_as("nanometer"), 180.0, 0.0],
            ),
        )


def _create_virtual_site_object(
    virtual_site_key: BaseVirtualSiteKey,
    virtual_site_potential,
    # interchange: "Interchange",
    # non_bonded_force: openmm.NonbondedForce,
) -> _VirtualSite:
    orientations = virtual_site_key.orientation_atom_indices

    # this check is basically "is this a SMIRNOFF virtual site?"
    # see commment at its class definition
    if isinstance(virtual_site_key, SMIRNOFFVirtualSiteKey):
        if virtual_site_key.type == "BondCharge":
            return _BondChargeVirtualSite(
                type="BondCharge",
                distance=virtual_site_potential.parameters["distance"],
                orientations=orientations,
            )
        elif virtual_site_key.type == "MonovalentLonePair":
            return _MonovalentLonePairVirtualSite(
                type="MonovalentLonePair",
                distance=virtual_site_potential.parameters["distance"],
                out_of_plane_angle=virtual_site_potential.parameters["outOfPlaneAngle"],
                in_plane_angle=virtual_site_potential.parameters["inPlaneAngle"],
                orientations=orientations,
            )
        elif virtual_site_key.type == "DivalentLonePair":
            return _DivalentLonePairVirtualSite(
                type="DivalentLonePair",
                distance=virtual_site_potential.parameters["distance"],
                out_of_plane_angle=virtual_site_potential.parameters["outOfPlaneAngle"],
                orientations=orientations,
            )
        elif virtual_site_key.type == "TrivalentLonePair":
            return _TrivalentLonePairVirtualSite(
                type="TrivalentLonePair",
                distance=virtual_site_potential.parameters["distance"],
                orientations=orientations,
            )

        else:
            raise NotImplementedError(virtual_site_key.type)

    # TODO: This case shouldn't be in openff/interchange/smirnoff!
    elif isinstance(virtual_site_key, ImportedVirtualSiteKey):
        if virtual_site_key.type == "ThreeParticleAverageSite":
            return _ThreeParticleAverageSite(  # type: ignore[return-value]
                particles=virtual_site_key.orientation_atom_indices,
                weights=virtual_site_potential.parameters["weights"],
            )

        else:
            raise NotImplementedError(virtual_site_key.type)

    else:
        raise NotImplementedError(virtual_site_key.type)


def _build_local_coordinate_frames(
    interchange,
    virtual_site_collection: SMIRNOFFVirtualSiteCollection,
) -> numpy.ndarray:
    """
    Build local coordinate frames.

    Adapted from an implementation in OpenFF Recharge (see `LICENSE-3RD-PARTY`).

    See Also
    --------
    https://github.com/openforcefield/openff-recharge/blob/0.5.0/openff/recharge/charges/vsite.py#L584

    """
    stacked_frames: list[list] = [[], [], [], []]

    for virtual_site_key, potential_key in virtual_site_collection.key_map.items():
        virtual_site = _create_virtual_site_object(
            virtual_site_key=virtual_site_key,
            virtual_site_potential=virtual_site_collection.potentials[potential_key],
        )

        # positions of all "orientation" atoms, not just the single "parent"
        orientation_coordinates = interchange.positions[
            virtual_site_key.orientation_atom_indices,
            :,
        ].m_as("nanometer")

        local_frame_weights = virtual_site.local_frame_weights

        weighted_coordinates = local_frame_weights @ orientation_coordinates

        origin = weighted_coordinates[0, :]

        xy_plane = weighted_coordinates[1:, :]

        xy_plane_norm = xy_plane / numpy.sqrt(
            (xy_plane * xy_plane).sum(-1),
        ).reshape(-1, 1)

        x_hat = xy_plane_norm[0, :]
        z_hat = numpy.cross(x_hat, xy_plane_norm[1, :])
        z_norm = numpy.linalg.norm(z_hat)

        THRESHOLD = 1e-6
        if z_norm > THRESHOLD:
            z_hat /= z_norm
        else:
            # Degenerate case for co-linear atoms, most likely as BondCharges
            # Construct an arbitrary perpendicular vector to x_hat
            if abs(x_hat[0]) < 0.9:  # corresponds to >= ~26 degrees away from x-axis
                z_hat = numpy.cross(x_hat, numpy.array([1.0, 0.0, 0.0]))
            else:
                z_hat = numpy.cross(x_hat, numpy.array([0.0, 1.0, 0.0]))
            z_hat = z_hat / numpy.linalg.norm(z_hat)

        y_hat = numpy.cross(z_hat, x_hat)

        stacked_frames[0].append(origin.reshape(1, -1))
        stacked_frames[1].append(x_hat.reshape(1, -1))
        stacked_frames[2].append(y_hat.reshape(1, -1))
        stacked_frames[3].append(z_hat.reshape(1, -1))

    local_frames = numpy.stack([numpy.vstack(frames) for frames in stacked_frames])

    return Quantity(local_frames, "nanometer")


def _convert_local_coordinates(
    local_frame_coordinates: numpy.ndarray,
    local_coordinate_frames: numpy.ndarray,
) -> numpy.ndarray:
    d = local_frame_coordinates[:, 0].reshape(-1, 1)

    theta = (local_frame_coordinates[:, 1] * _DEGREES_TO_RADIANS).reshape(-1, 1)
    phi = (local_frame_coordinates[:, 2] * _DEGREES_TO_RADIANS).reshape(-1, 1)

    cos_theta = numpy.cos(theta)
    sin_theta = numpy.sin(theta)

    cos_phi = numpy.cos(phi)
    sin_phi = numpy.sin(phi)

    # Here we use cos(phi) in place of sin(phi) and sin(phi) in place of cos(phi)
    # this is because we want phi=0 to represent a 0 degree angle from the x-y plane
    # rather than 0 degrees from the z-axis.
    vsite_positions = local_coordinate_frames[0] + d * (
        cos_theta * cos_phi * local_coordinate_frames[1]
        + sin_theta * cos_phi * local_coordinate_frames[2]
        + sin_phi * local_coordinate_frames[3]
    )

    return vsite_positions


def _generate_positions(
    interchange,
    virtual_site_collection: SMIRNOFFVirtualSiteCollection,
    conformer: _Quantity | None = None,
) -> Quantity:
    # TODO: Capture these objects instead of generating them on-the-fly so many times
    try:
        local_frame_coordinates = numpy.vstack(
            [
                _create_virtual_site_object(
                    virtual_site_key,
                    virtual_site_collection.potentials[potential_key],
                ).local_frame_coordinates
                for virtual_site_key, potential_key in virtual_site_collection.key_map.items()
            ],
        )

        local_coordinate_frames = _build_local_coordinate_frames(
            interchange,
            virtual_site_collection,
        )

        virtual_site_positions = _convert_local_coordinates(
            local_frame_coordinates,
            local_coordinate_frames,
        )

        return Quantity(virtual_site_positions, "nanometer")

    except AttributeError:
        # Handle case of virtual sites imported from OpenMM
        # TODO: Handle case of mixed SMIRNOFF and OpenMM virtual sites
        _virtual_site_positions: list[Quantity] = list()

        for key, val in virtual_site_collection.key_map.items():
            atom_indices = key.orientation_atom_indices
            weights = virtual_site_collection.potentials[val].parameters["weights"].m

            assert len(atom_indices) == len(weights)

            _virtual_site_positions.append(
                sum(interchange.positions[atom_indices[i]] * weights[i] for i in range(len(weights))),
            )

        return numpy.vstack(_virtual_site_positions)



"""Custom exceptions used in Interchange."""


class InterchangeException(Exception):
    """Base class for all Interchange exceptions."""


class SMIRNOFFParameterAttributeNotImplementedError(InterchangeException):
    """
    Exception for when a parameter attribute is supported by the SMIRNOFF specification but not yet implemented.

    For example, this was raised when k_bondorder (used in bond order-based interpolation of force constants)
    before the behavior was supported.
    """


class SMIRNOFFHandlersNotImplementedError(InterchangeException):
    """
    Exception for when some parameter handlers in the SMIRNOFF specification are not implemented here.
    """


class SMIRNOFFVersionNotSupportedError(InterchangeException):
    """
    Exception for when a parameter handler's version is not supported.
    """


class InvalidParameterHandlerError(InterchangeException, ValueError):
    """
    Generic exception for mismatch between expected and found ParameterHandler types.
    """


class InvalidBoxError(InterchangeException, ValueError):
    """
    Generic exception for errors reading box data.
    """


class InvalidPositionsError(InterchangeException, ValueError):
    """
    Generic exception for errors with positions.
    """


class NoPositionsError(InterchangeException, NotImplementedError):
    """
    Exception for trying to access or set positions which are not stored.
    """


class InvalidTopologyError(InterchangeException, ValueError):
    """
    Generic exception for errors reading chemical topology data.
    """


class NonbondedEnergyError(InterchangeException, AssertionError):
    """
    Exception for when non-bonded energies computed from different objects differ.
    """


class InvalidExpressionError(InterchangeException, ValueError):
    """
    Exception for when an expression cannot safely be interpreted.
    """


class UnsupportedCutoffMethodError(InterchangeException):
    """
    Exception for a cutoff method that is invalid or not supported by an engine.
    """


class UnimplementedCutoffMethodError(InterchangeException):
    """
    Exception for a cutoff method that should be supported but it not yet implemented.
    """


class UnsupportedMixingRuleError(InterchangeException):
    """
    Raised when attempting to use a mixing rule is invalid, unsupported or otherwise not implemented.
    """


class UnsupportedParameterError(InterchangeException, ValueError):
    """
    Exception for parameters having unsupported values, i.e. non-1.0 idivf.
    """


class UnsupportedBoxError(InterchangeException, ValueError):
    """
    Exception for processing an unsupported box, probably non-orthogonal.
    """


class UnsupportedImportError(InterchangeException):
    """
    Generic exception for attempting to import from an unsupported data format.
    """


class UnsupportedExportError(InterchangeException):
    """
    Exception for attempting to write to an unsupported file format.
    """


class NonperiodicNoCutoffNotSupportedError(UnsupportedExportError):
    """
    Exception for when an engine does not support "no-cutoff" non-bonded methods in non-periodic system.
    """

    def __init__(self, message):
        super().__init__(
            message
            + "\nFor more see: https://docs.openforcefield.org/projects/interchange/en/stable/using/edges.html#no-cutoff-support",
        )


class UnsupportedCombinationError(InterchangeException):
    """General exception for something going wrong in Interchange object combination."""


class CutoffMismatchError(UnsupportedCombinationError):
    """Non-bonded cutoffs do not match."""


class SwitchingFunctionMismatchError(UnsupportedCombinationError):
    """Switching distances or the use of a switching function does not match."""


class PluginCompatibilityError(InterchangeException):
    """A plugin is incompatible with the current version of Interchange in the way it is called."""


class CannotSetSwitchingFunctionError(InterchangeException):
    """
    Unable to set a switching function.
    """


class CannotInferEnergyError(InterchangeException):
    """
    Failed to infer a physical interpretation of this energy term.
    """


class CannotInferNonbondedEnergyError(CannotInferEnergyError):
    """
    Failed to infer a physical interpretation of this non-bonded energy.
    """


class InvalidWriterError(InterchangeException):
    """
    An unknown or unimplemnted writer was specified.
    """


class ConversionError(InterchangeException):
    """
    Base exception for error handling during object conversion.
    """


class MissingBoxError(InterchangeException):
    """
    Exception for when box vectors are needed but missing.
    """


class MissingPositionsError(InterchangeException):
    """
    Exception for when positions are needed but missing.
    """


class MissingParameterHandlerError(InterchangeException):
    """
    Exception for when a parameter handler is requested but not found.
    """


class MissingParametersError(InterchangeException):
    """
    Exception for when parameters are needed but missing.
    """


class MissingBondOrdersError(InterchangeException):
    """
    Exception for when a parameter handler needs fractional bond orders but they are missing.
    """


class DuplicateMoleculeError(InterchangeException, ValueError):
    """
    Exception for when molecules are not unique.
    """


class MissingUnitError(InterchangeException, ValueError):
    """
    Exception for data missing a unit tag.
    """


class UnitValidationError(InterchangeException, ValueError):
    """
    Exception for bad behavior when validating unit-tagged data.
    """


class NonbondedCompatibilityError(InterchangeException):
    """
    Exception for unsupported combination of nonbonded methods.
    """


class MissingNonbondedCompatibilityError(InterchangeException):
    """
    Exception for uncovered combination of nonbonded methods.
    """


class InternalInconsistencyError(InterchangeException):
    """
    Fallback exception for bad behavior releating to a self-inconsistent internal state.

    These should not be reached but are raised to safeguard against problematic edge cases silently passing.
    """


class AmberError(InterchangeException):
    """
    Base exception for handling Amber-related errors.
    """


class AmberExecutableNotFoundError(AmberError):
    """
    Exception for when an Amber-related excutable is not found.
    """


class SanderError(AmberError):
    """
    Exception for when a sander subprocess fails.
    """


class GMXError(InterchangeException):
    """
    Exception for when a GROMACS subprocess fails.
    """


class GMXNotFoundError(GMXError):
    """
    Exception for when no GROMACS executable is found.
    """


class GMXGromppError(GMXError):
    """
    Exception for when `gmx grompp` fails.
    """


class GMXMdrunError(GMXError):
    """
    Exception for when `gmx mdrun` fails.
    """


class LAMMPSError(InterchangeException):
    """
    Base exception for handling LAMMPS-related errors.
    """


class LAMMPSNotFoundError(LAMMPSError):
    """
    Exception for when no LAMMPS executable is found.
    """


class LAMMPSRunError(LAMMPSError):
    """
    Exception for when a LAMMPS subprocess fails.
    """


class EnergyError(InterchangeException):
    """
    Base class for energies in reports not matching.
    """


class InvalidEnergyError(InterchangeException):
    """
    Energy type not understood.
    """


class IncompatibleTolerancesError(InterchangeException):
    """
    Exception for when one report has a value for an energy group but the other does not.
    """


class NegativeMassError(InterchangeException):
    """Exception for when a particle mass is negative."""


class MissingVirtualSitesError(InterchangeException):
    """
    Raise when virtual sites are expected to exist but are not found.
    """


class VirtualSiteTypeNotImplementedError(InterchangeException):
    """
    Raised when this type of virtual site is not yet implemented.
    """


class NonIntegralMoleculeChargeError(InterchangeException):
    """
    Exception raised when the partial charges on a molecule do not sum up to its formal charge.
    """


class MissingPartialChargesError(InterchangeException):
    """
    A molecule is missing partial charges.
    """


class PresetChargesError(InterchangeException):
    """
    The `charge_from_molecules` is structured in an undefined way.
    """


class UnassignedValenceError(InterchangeException):
    """Exception raised when there are valence terms for which a ParameterHandler did not find matches."""


class UnassignedBondError(UnassignedValenceError):
    """Exception raised when there are bond terms for which a ParameterHandler did not find matches."""


class UnassignedAngleError(UnassignedValenceError):
    """Exception raised when there are angle terms for which a ParameterHandler did not find matches."""


class UnassignedTorsionError(UnassignedValenceError):
    """Exception raised when there are torsion terms for which a ParameterHandler did not find matches."""


class MissingValenceError(InterchangeException):
    """Exception raised when there are valence interactions for which no parameters are found."""


class MissingBondError(UnassignedValenceError):
    """Exception raised when there exists a bond for which no parameters are found."""


class MissingAngleError(UnassignedValenceError):
    """Exception raised when there exists an angle for which no parameters are found."""


class MissingTorsionError(UnassignedValenceError):
    """Exception raised when there exists a torsion for which no parameters are found."""


class PACKMOLRuntimeError(InterchangeException):
    """Exception raised when PACKMOL fails to execute / converge."""


class PACKMOLValueError(InterchangeException):
    """Exception raised when a bad input is passed to a PACKMOL wrapper."""


class MinimizationError(InterchangeException):
    """Exception raised when an energy minimization fails to converge."""


class ExperimentalFeatureException(InterchangeException):
    """Exception raised when an experimental feature is used without opt-in."""


class GROMACSParseError(InterchangeException, ValueError):
    """Exception raised when a GROMACS file cannot be parsed correctly."""



from typing import Any

class TopologyGraph:
    @classmethod
    def from_openff_topology(cls: Any, openff_topology: Any) -> Any: ...
    def add_atom(
        self,
        index: int,
        name: str,
        atomic_number: int | None = None,
        element: Any | None = None,
        **kwargs,
    ) -> None: ...
    def add_bond(self, atom_1_index: int, atom_2_index: int) -> None: ...
    @classmethod
    def from_parmed(cls: Any, structure: Any) -> Any: ...



import pytest
from openff.toolkit import ForceField, Quantity

from openff.interchange._tests import get_test_file_path
from openff.interchange.constants import kj_mol
from openff.interchange.drivers.openmm import _process, get_openmm_energies

openmm = pytest.importorskip("openmm")


def test_detailed_requires_split_nonbonded_forces(methane_dimer):
    """Ensure that a detailed report requires non-bonded forces to be split."""
    with pytest.raises(ValueError, match="split out into different forces"):
        get_openmm_energies(methane_dimer, combine_nonbonded_forces=True, detailed=True)


def test_key_order(methane_dimer):
    """Ensure that the keys in the report are in a consistent order."""
    report = get_openmm_energies(methane_dimer, detailed=True, combine_nonbonded_forces=False)

    expected_keys = [
        "Bond",
        "Angle",
        "Torsion",
        "RBTorsion",
        "vdW",
        "vdW 1-4",
        "Electrostatics",
        "Electrostatics 1-4",
    ]

    assert [*report.energies.keys()] == [key for key in expected_keys if key in report.energies]


class TestProcess:
    @pytest.fixture
    def dummy_system(self):
        system = openmm.System()
        for force in [
            openmm.PeriodicTorsionForce,
            openmm.HarmonicAngleForce,
            openmm.HarmonicBondForce,
            openmm.RBTorsionForce,
            openmm.NonbondedForce,
        ]:
            system.addForce(force())

        return system

    @pytest.fixture
    def dummy_system_split(self, dummy_system):
        dummy_system.addForce(openmm.CustomNonbondedForce("sigma*epsilon"))
        dummy_system.addForce(openmm.CustomBondForce("sigma*epsilon"))
        dummy_system.addForce(openmm.CustomBondForce("qq"))

        return dummy_system

    def test_simple(self, dummy_system):
        processed = _process(
            {
                0: 1.1 * kj_mol,
                1: 2.2 * kj_mol,
                2: 3.3 * kj_mol,
                3: 4.4 * kj_mol,
                4: 5.5 * kj_mol,
            },
            dummy_system,
            True,
            False,
        )

        assert processed["Bond"].m_as(kj_mol) == 3.3
        assert processed["Angle"].m_as(kj_mol) == 2.2
        assert processed["Torsion"].m_as(kj_mol) == 1.1
        assert processed["RBTorsion"].m_as(kj_mol) == 4.4
        assert processed["Nonbonded"].m_as(kj_mol) == 5.5

    def test_split_forces(self, dummy_system_split):
        processed = _process(
            {
                0: 1.1 * kj_mol,
                1: 2.2 * kj_mol,
                2: 3.3 * kj_mol,
                3: 4.4 * kj_mol,
                4: 0.5 * kj_mol,
                5: -1 * kj_mol,
                6: -2 * kj_mol,  # vdW
                7: -3 * kj_mol,  # Electrostatics
            },
            dummy_system_split,
            False,
            True,
        )

        assert processed["Electrostatics"].m_as(kj_mol) == 0.5
        assert processed["Electrostatics 1-4"].m_as(kj_mol) == -3
        assert processed["vdW"].m_as(kj_mol) == -1
        assert processed["vdW 1-4"].m_as(kj_mol) == -2


class TestReportWithPlugins:
    @pytest.fixture
    def de_force_field(self) -> ForceField:
        pytest.importorskip("smirnoff_plugins")
        pytest.importorskip("openeye")
        return ForceField(
            get_test_file_path("de-force-1.0.1.offxml"),
            load_plugins=True,
        )

    @pytest.mark.parametrize("detailed", [True, False])
    def test_nonzero_vdw(self, ligand, de_force_field, detailed):
        energies = get_openmm_energies(
            de_force_field.create_interchange(ligand.to_topology()),
            combine_nonbonded_forces=False,
            detailed=detailed,
        )

        assert energies["vdW"].m != 0

        if detailed:
            assert energies["vdW 1-4"].m != 0

    def test_detailed_same_total_energy(self, ligand, de_force_field):
        assert get_openmm_energies(
            de_force_field.create_interchange(ligand.to_topology()),
            combine_nonbonded_forces=False,
            detailed=True,
        ).total_energy.m == pytest.approx(
            get_openmm_energies(
                de_force_field.create_interchange(ligand.to_topology()),
                combine_nonbonded_forces=False,
                detailed=False,
            ).total_energy.m,
        )

    def test_intermolecular_plugin_vdw_energies_reported(
        self,
        de_force_field,
        methane_dimer,
    ):
        new_interchange = de_force_field.create_interchange(methane_dimer.topology)
        new_interchange.positions = methane_dimer.positions

        new_interchange.minimize(
            engine="openmm",
            force_tolerance=Quantity(0.01, "kilojoule_per_mole / nanometer"),
        )

        energies = get_openmm_energies(
            new_interchange,
            combine_nonbonded_forces=False,
            detailed=True,
        )

        # vdW (non-LJ) interaction energy should be slightly negative; for comparison,
        # sage number is -1.377 kJ/mol
        assert -5 < energies["vdW"].m_as(kj_mol) < -1



#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/stable/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# Incase the project was not installed
import importlib
import os

from openff.interchange import __version__

# -- Project information -----------------------------------------------------

project = "openff-interchange"
copyright = (
    "2020, Open Force Field Initiative. Project structure based on the "
    "Computational Molecular Science Python Cookiecutter version 1.2"
)
author = "Open Force Field Initiative"
version = __version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.autosummary",
    "sphinx.ext.todo",
    "openff_sphinx_theme",
    "sphinxcontrib.autodoc_pydantic",
    "sphinxcontrib.mermaid",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "openff.toolkit": (
        "https://open-forcefield-toolkit.readthedocs.io/en/stable/",
        None,
    ),
    "mdtraj": ("https://www.mdtraj.org/1.9.5/", None),
    "openff.docs": (
        "https://docs.openforcefield.org/en/latest/",
        None,
    ),
    "setuptools": ("https://setuptools.pypa.io/en/latest/", None),
    "openmm": ("http://docs.openmm.org/latest/api-python/", None),
}

mermaid_init_js = """mermaid.initialize({
    startOnLoad:true,
    fontFamily:"montserrat,sans",
    theme: "default",
});"""

## Suppress warnings about ambiguous Python references
suppress_warnings = [
    "ref.python",
]

autosummary_generate = True
autosummary_imported_members = False
autosummary_context = {
    # Modules to exclude from API docs
    "exclude_modules": [
        "openff.interchange._tests",  # Maybe this is excluded by default?
    ],
}
autosummary_ignore_module_all = False
napoleon_google_docstring = False
napoleon_use_param = True
napoleon_use_ivar = True
napoleon_use_rtype = False

autodoc_mock_imports = ["jax"]
autodoc_default_options = {
    "member-order": "bysource",
    "undoc-members": True,
    "inherited-members": "",
    "show-inheritance": True,
}
autodoc_preserve_defaults = True
autodoc_inherit_docstrings = False
autodoc_typehints = "both"
autodoc_typehints_format = "short"
autodoc_typehints_description_target = "documented"

# Workaround for autodoc_typehints_format not working for attributes
# see https://github.com/sphinx-doc/sphinx/issues/10290#issuecomment-1079740009
python_use_unqualified_type_names = True

# autodoc_pydantic settings
autodoc_pydantic_show_config = False
autodoc_pydantic_model_show_config = False
autodoc_pydantic_model_show_config_summary = False
autodoc_pydantic_show_validators = False
autodoc_pydantic_model_show_validators = False
autodoc_pydantic_field_show_alias = False
autodoc_pydantic_model_show_json = False

myst_enable_extensions = [
    "deflist",
    "smartquotes",
    "replacements",
    "dollarmath",
    "colon_fence",
]

# sphinx-notfound-page
# https://github.com/readthedocs/sphinx-notfound-page
# Renders a 404 page with absolute links

if importlib.util.find_spec("notfound"):
    extensions.append("notfound.extension")

    notfound_context = {
        "title": "404: File Not Found",
        "body": """
    <h1>404: File Not Found</h1>
    <p>
        Sorry, we couldn't find that page. This often happens as a result of
        following an outdated link. Please check the
        <a href="https://openff-interchange.readthedocs.io/en/stable/">latest stable version</a>
        of the docs, unless you're sure you want an earlier version, and
        try using the search box or the navigation menu on the left.
    </p>
    <p>
    </p>
    """,
    }

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "README.md"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "default"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "openff_sphinx_theme"
html_sidebars = {"**": ["globaltoc.html", "localtoc.html", "searchbox.html"]}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    # Repository integration
    # Set the repo url for the link to appear
    "repo_url": "https://github.com/openforcefield/openff-interchange",
    # The name of the repo. If must be set if repo_url is set
    "repo_name": "openff-interchange",
    # Must be one of github, gitlab or bitbucket
    "repo_type": "github",
    # Colour for sidebar captions and other accents. One of
    # openff-blue, openff-toolkit-blue, openff-dataset-yellow,
    # openff-evaluator-orange, aquamarine, lilac, amaranth, grape,
    # violet, pink, pale-green, green, crimson, eggplant, turquoise,
    # or a tuple of three ints in the range [0, 255] corresponding to
    # a position in RGB space.
    "color_accent": "pink",
    # Content Minification for deployment, prettification for debugging
    "html_minify": False,  # q for mermaid to work
    "css_minify": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "interchangedoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "interchange.tex",
        "Interchange Documentation",
        "interchange",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, "interchange", "Interchange Documentation", [author], 1),
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "interchange",
        "Interchange Documentation",
        author,
        "interchange",
        "A molecular interchange object from the Open Force Field Initiative",
        "Miscellaneous",
    ),
]


# -- ReadTheDocs configuration -------------------------------------------------

if os.environ.get("READTHEDOCS", "") == "True":
    # Define the canonical URL if you are using a custom domain on Read the Docs
    html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "")

    # Tell Jinja2 templates the build is running on Read the Docs
    if "html_context" not in globals():
        html_context = {}
    html_context["READTHEDOCS"] = True



from openff.toolkit import Quantity
from openff.utilities.testing import skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer, needs_gmx
from openff.interchange.drivers.gromacs import get_gromacs_energies
from openff.interchange.drivers.openmm import get_openmm_energies


@skip_if_missing("openmm")
@needs_gmx
def test_group_impropers(cb8_host, no_charges):
    out = Interchange.from_smirnoff(no_charges, [cb8_host], box=[4, 4, 4])

    openmm_torsions = get_openmm_energies(out)["Torsion"]
    gromacs_torsions = get_gromacs_energies(out)["Torsion"]

    assert abs(openmm_torsions - gromacs_torsions).m_as("kilojoule_per_mole") < 1e-3


@needs_gmx
def test_14_in_detailed_report(sage):
    """Ensure that 1-4 interactions are included in the detailed report and non-zero."""
    octanol = MoleculeWithConformer.from_smiles("CCCCCCCC")
    topology = octanol.to_topology()
    topology.box_vectors = Quantity([3, 3, 3], "nanometer")

    report = get_gromacs_energies(sage.create_interchange(topology), detailed=True)

    assert report["vdW 1-4"].m != 0.0
    assert report["Electrostatics 1-4"].m != 0.0


@needs_gmx
def test_key_order(methane_dimer):
    """Ensure that the keys in the report are in a consistent order."""
    report = get_gromacs_energies(methane_dimer, detailed=True)

    expected_keys = [
        "Bond",
        "Angle",
        "Torsion",
        "RBTorsion",
        "vdW",
        "vdW 1-4",
        "Electrostatics",
        "Electrostatics 1-4",
    ]

    assert [*report.energies.keys()] == [key for key in expected_keys if key in report.energies]



"""Functions for running energy evluations with LAMMPS."""

import tempfile

import numpy
from openff.toolkit import Quantity
from openff.utilities import MissingOptionalDependencyError, requires_package

from openff.interchange import Interchange
from openff.interchange.drivers.report import EnergyReport
from openff.interchange.exceptions import LAMMPSNotFoundError, LAMMPSRunError


def get_lammps_energies(
    interchange: Interchange,
    round_positions: int | None = None,
    detailed: bool = False,
) -> EnergyReport:
    """
    Given an OpenFF Interchange object, return single-point energies as computed by LAMMPS.

    Parameters
    ----------
    interchange
        An OpenFF Interchange object to compute the single-point energy of
    round_positions
        The number of decimal places, in nanometers, to round positions. This can be useful when
        comparing to i.e. GROMACS energies, in which positions may be rounded.
    detailed
        If True, return a detailed energy report containing all energy components.

    Returns
    -------
    report
        An `EnergyReport` object containing the single-point energies.

    """
    try:
        return _process(
            _get_lammps_energies(interchange, round_positions),
            detailed,
        )
    except MissingOptionalDependencyError:
        raise LAMMPSNotFoundError


@requires_package("lammps")
def _get_lammps_energies(
    interchange: Interchange,
    round_positions: int | None = None,
) -> dict[str, Quantity]:
    import lammps

    if round_positions is not None:
        interchange.positions = numpy.round(
            interchange.positions,  # type: ignore[arg-type]
            round_positions,
        )

    with tempfile.TemporaryDirectory():
        interchange.to_lammps("out")

    # By default, LAMMPS spits out logs to the screen, turn it off
    # https://matsci.org/t/how-to-remove-or-redirect-python-lammps-stdout/38075/5
    # not that this is not sent to STDOUT, so `contextlib.redirect_stdout` won't work
    runner = lammps.lammps(cmdargs=["-screen", "none", "-nocite"])

    try:
        runner.file("out_pointenergy.in")
    # LAMMPS does not raise a custom exception :(
    except Exception as error:
        raise LAMMPSRunError from error

    # thermo_style custom ebond eangle edihed eimp epair evdwl ecoul elong etail pe
    parsed_energies = [Quantity(energy, "kilocalorie_per_mole") for energy in runner.last_thermo().values()]

    # TODO: Sanely map LAMMPS's energy names to the ones we care about
    return {
        "Bond": parsed_energies[0],
        "Angle": parsed_energies[1],
        "ProperTorsion": parsed_energies[2],
        "ImproperTorsion": parsed_energies[3],
        "vdW": parsed_energies[5],
        "DispersionCorrection": parsed_energies[8],
        "ElectrostaticsShort": parsed_energies[6],
        "ElectrostaticsLong": parsed_energies[7],
    }


def _process(
    energies: dict[str, Quantity],
    detailed: bool = False,
) -> EnergyReport:
    if detailed:
        return EnergyReport(energies=energies)

    return EnergyReport(
        energies={
            "Bond": energies["Bond"],
            "Angle": energies["Angle"],
            "Torsion": energies["ProperTorsion"] + energies["ImproperTorsion"],
            "vdW": energies["vdW"] + energies["DispersionCorrection"],
            "Electrostatics": (energies["ElectrostaticsShort"] + energies["ElectrostaticsLong"]),
        },
    )



"""Functions for running energy evluations with Amber."""

import subprocess
import tempfile
from pathlib import Path
from shutil import which

from openff.toolkit import Quantity
from openff.utilities.utilities import temporary_cd

from openff.interchange import Interchange
from openff.interchange.components.mdconfig import MDConfig
from openff.interchange.drivers.report import EnergyReport
from openff.interchange.exceptions import (
    AmberError,
    AmberExecutableNotFoundError,
    SanderError,
)


def get_amber_energies(
    interchange: Interchange,
    detailed: bool = False,
) -> EnergyReport:
    """
    Given an OpenFF Interchange object, return single-point energies as computed by Amber.

    Parameters
    ----------
    interchange
        An OpenFF Interchange object to compute the single-point energy of
    detailed
        If True, return a detailed report containing the energies of each

    Returns
    -------
    report
        An `EnergyReport` object containing the single-point energies.

    """
    return _process(
        _get_amber_energies(
            interchange=interchange,
        ),
        detailed=False,
    )


def _get_amber_energies(
    interchange: Interchange,
) -> dict[str, Quantity]:
    with tempfile.TemporaryDirectory() as tmpdir:
        with temporary_cd(tmpdir):
            # TODO: Use to_amber if implemented
            interchange.to_inpcrd("out.inpcrd")
            interchange.to_prmtop("out.prmtop")

            mdconfig = MDConfig.from_interchange(interchange)
            mdconfig.write_sander_input_file("run.in")

            return _run_sander(
                prmtop_file="out.prmtop",
                inpcrd_file="out.inpcrd",
                input_file="run.in",
            )


def _run_sander(
    inpcrd_file: Path | str,
    prmtop_file: Path | str,
    input_file: Path | str,
) -> dict[str, Quantity]:
    """
    Given Amber files, return single-point energies as computed by Amber.

    Parameters
    ----------
    prmtop_file
        The path to an Amber topology (`.prmtop`) file.
    inpcrd_file
        The path to an Amber coordinate (`.inpcrd`) file.
    input_file
        The path to an Amber/sander input (`.in`) file.

    Returns
    -------
    energies
        A dictionary of energies, keyed by the GROMACS energy term name.

    """
    if not which("sander"):
        raise AmberExecutableNotFoundError(
            "Unable to find the 'sander' executable. Please ensure that "
            "the Amber executables are installed and in your PATH.",
        )

    sander_cmd = f"sander -i {input_file} -c {inpcrd_file} -p {prmtop_file} -o out.mdout -O"

    sander = subprocess.Popen(
        sander_cmd,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )

    _, err = sander.communicate()

    if sander.returncode:
        raise SanderError(err)

    return _parse_amber_energy("mdinfo")


def _parse_amber_energy(mdinfo: str) -> dict[str, Quantity]:
    """
    Parse AMBER output file and group the energy terms in a dict.

    This code is partially copied from InterMol, see
    https://github.com/shirtsgroup/InterMol/tree/v0.1/intermol/amber/
    """
    with open(mdinfo) as f:
        all_lines = f.readlines()

    # Find where the energy information starts.
    for i, line in enumerate(all_lines):
        # Seems to hit energy minimization
        if line[0:8] == "   NSTEP":
            startline = i + 2
            break
        # Seems to hit MD "runs"
        elif line[0:6] == " NSTEP":
            startline = i
            break
    else:
        raise AmberError(
            f"Unable to detect where energy info starts in AMBER output file: {mdinfo}",
        )

    # Strange ranges for amber file data.
    ranges = [[1, 24], [26, 49], [51, 77]]

    e_out = dict()
    potential = Quantity(0, "kilocalories_per_mole")
    for line in all_lines[startline + 1 :]:
        if "=" in line:
            for i in range(3):
                r = ranges[i]
                term = line[r[0] : r[1]]
                if "=" in term:
                    energy_type, energy_value = term.strip().split("=")
                    try:
                        energy_value = Quantity(float(energy_value), "kilocalories_per_mole")
                    except ValueError as error:
                        raise AmberError(
                            f"Found bad energy value '{energy_value}' associated with energy type '{energy_type}'",
                        ) from error
                    potential += energy_value
                    energy_type = energy_type.rstrip()
                    e_out[energy_type] = energy_value
        else:
            break
    e_out["ENERGY"] = potential

    return e_out


def _get_amber_energy_vdw(amber_energies: dict) -> Quantity:
    """Get the total nonbonded energy from a set of Amber energies."""
    amber_vdw = Quantity(0, "kilojoule_per_mole")
    for key in ["VDWAALS", "1-4 VDW", "1-4 NB"]:
        if key in amber_energies:
            amber_vdw += amber_energies[key]

    return amber_vdw


def _get_amber_energy_coul(amber_energies: dict) -> Quantity:
    """Get the total nonbonded energy from a set of Amber energies."""
    amber_coul = Quantity(0, "kilojoule_per_mole")
    for key in ["EEL", "1-4 EEL"]:
        if key in amber_energies:
            amber_coul += amber_energies[key]

    return amber_coul


def _process(
    energies: dict[str, Quantity],
    detailed: bool = False,
) -> EnergyReport:
    if detailed:
        return EnergyReport(energies=energies)

    return EnergyReport(
        energies={
            "Bond": energies["BOND"],
            "Angle": energies["ANGLE"],
            "Torsion": energies["DIHED"],
            "vdW": _get_amber_energy_vdw(energies),
            "Electrostatics": _get_amber_energy_coul(energies),
        },
    )



import shutil

import pytest
from openff.utilities import has_executable, has_package


@pytest.mark.skipif(
    not has_executable("packmol"),
    reason="Packmol is not available on this system",
)
def test_basic_reexport():
    """Test that the packmol module can be imported."""
    from openff.interchange.components._packmol import _find_packmol, pack_box

    assert _find_packmol() == shutil.which("packmol")

    assert "interchange" not in pack_box.__module__


@pytest.mark.skipif(
    has_package("openff.packmol"),
    reason="openff.packmol is not installed",
)
def test_import_error_when_openff_packmol_not_installed():
    with pytest.raises(
        ImportError,
    ):
        from openff.interchange.components._packmol import pack_box

        assert "interchange" not in pack_box.__module__



"""Runtime settings for MD simulations."""

import warnings
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Literal

from openff.toolkit import Quantity
from pydantic import Field

from openff.interchange._annotations import _DistanceQuantity
from openff.interchange.components.toolkit import _get_num_h_bonds
from openff.interchange.constants import _PME
from openff.interchange.exceptions import (
    NonperiodicNoCutoffNotSupportedError,
    UnsupportedCutoffMethodError,
    UnsupportedExportError,
)
from openff.interchange.pydantic import _BaseModel
from openff.interchange.warnings import SwitchingFunctionNotImplementedWarning

if TYPE_CHECKING:
    from openff.interchange import Interchange

MDP_HEADER = """\
; Single point energy
nsteps                   = 0      ; Perform no MD steps
nstenergy                = 1000   ; Write energies to .EDR file
continuation             = yes    ; Do not apply constraints at start
cutoff-scheme            = verlet ; Use a Verlet pair list for cutoffs

; Force field configuration
disp-corr                = Ener
"""


class MDConfig(_BaseModel):
    """A partial superset of runtime configurations for MD engines."""

    periodic: bool = Field(
        True,
        description="Whether or not the system is periodic.",
    )
    constraints: str = Field(
        "none",
        description="The type of constraints to be used in the simulation.",
    )
    vdw_method: Literal["cutoff", "pme", "no-cutoff"] = Field(
        "cutoff",
        description="The method used to calculate the vdW interactions.",
    )
    vdw_cutoff: _DistanceQuantity = Field(
        Quantity(9.0, "angstrom"),
        description="The distance at which pairwise interactions are truncated",
    )
    mixing_rule: str = Field(
        "lorentz-berthelot",
        description="The mixing rule (combination rule, combining rule) used in computing pairwise vdW interactions",
    )

    switching_function: bool = Field(
        False,
        description="Whether or not to use a switching function for the vdw interactions",
    )
    switching_distance: _DistanceQuantity = Field(
        Quantity(0.0, "angstrom"),
        description="The distance at which the switching function is applied",
    )
    coul_method: str = Field(
        "",
        description="The method used to compute pairwise electrostatic interactions",
    )
    coul_cutoff: _DistanceQuantity = Field(
        Quantity(9.0, "angstrom"),
        description=(
            "The distance at which electrostatic interactions are truncated or transition from short- to long-range."
        ),
    )

    @classmethod
    def from_interchange(cls, interchange: "Interchange") -> "MDConfig":
        """Generate a MDConfig object from an Interchange object."""
        mdconfig = cls(
            periodic=interchange.box is not None,
            constraints=_infer_constraints(interchange),
        )
        if "vdW" in interchange.collections:
            vdw_collection = interchange["vdW"]

            if interchange.box is None:
                mdconfig.vdw_method = vdw_collection.nonperiodic_method
            else:
                mdconfig.vdw_method = vdw_collection.periodic_method
                mdconfig.vdw_cutoff = vdw_collection.cutoff

            mdconfig.mixing_rule = vdw_collection.mixing_rule

            if vdw_collection.switch_width is not None:
                if vdw_collection.switch_width.m == 0:
                    mdconfig.switching_function = False
                else:
                    mdconfig.switching_function = True
                    mdconfig.switching_distance = mdconfig.vdw_cutoff - vdw_collection.switch_width
            else:
                mdconfig.switching_function = False

        if "Electrostatics" in interchange.collections:
            mdconfig.coul_method = getattr(
                interchange["Electrostatics"],
                "periodic_potential" if mdconfig.periodic else "nonperiodic_potential",
            )
            mdconfig.coul_cutoff = interchange["Electrostatics"].cutoff

        return mdconfig

    def apply(self, interchange: "Interchange"):
        """Attempt to apply these settings to an Interchange object."""
        if self.periodic:
            if interchange.box is None:
                interchange.box = Quantity([10, 10, 10], "nanometer")
        else:
            interchange.box = None

        if "vdW" in interchange.collections:
            vdw_collection = interchange["vdW"]

            if interchange.box is None:
                vdw_collection.nonperiodic_method = self.vdw_method
            else:
                vdw_collection.periodic_method = self.vdw_method

            vdw_collection.cutoff = self.vdw_cutoff
            vdw_collection.mixing_rule = self.mixing_rule

            if self.switching_function:
                vdw_collection.switch_width = self.vdw_cutoff - self.switching_distance
            else:
                vdw_collection.switch_width = Quantity(0.0, "angstrom")

        if "Electrostatics" in interchange.collections:
            electrostatics = interchange["Electrostatics"]
            if self.coul_method.lower() == "pme":
                electrostatics.periodic_potential = _PME  # type: ignore[assignment]
            else:
                electrostatics.periodic_potential = self.coul_method  # type: ignore[assignment]
            electrostatics.cutoff = self.coul_cutoff

    def write_mdp_file(self, mdp_file: str = "auto_generated.mdp") -> None:
        """
        Write a GROMACS `.mdp` file for running single-point energies.

        Parameters
        ----------
        mdp_file
            The name of the file to write the input to, defaults to `auto_generated.mdp`.
        """
        # Construct the MDP file in memory so nothing is written to disk if an
        # error is encountered.
        with StringIO() as mdp:
            mdp.write(MDP_HEADER)

            if self.periodic:
                mdp.write("pbc                      = xyz\n")
            else:
                mdp.write("pbc                      = no\n")

            mdp.write(f"constraints              = {self.constraints}\n")

            coul_cutoff = round(self.coul_cutoff.m_as("nanometer"), 4)

            if self.coul_method == "cutoff":
                mdp.write("coulombtype              = Cut-off\n")
                mdp.write("coulomb-modifier         = None\n")
                mdp.write(f"rcoulomb                 = {coul_cutoff}\n")
            elif self.coul_method in (_PME, "PME", "pme"):
                if not self.periodic:
                    raise UnsupportedCutoffMethodError(
                        "PME is not valid with a non-periodic system.",
                    )
                mdp.write("coulombtype              = PME\n")
                mdp.write(f"rcoulomb                 = {coul_cutoff}\n")
                mdp.write("coulomb-modifier         = None\n")
                mdp.write("fourier-spacing          = 0.12\n")
                # TODO: Wire this through like `ewald_tolerance` in `to_openmm`
                mdp.write("ewald-rtol               = 1e-4\n")
            elif self.coul_method == "reactionfield":
                mdp.write("coulombtype              = Reaction-field\n")
                mdp.write(f"rcoulomb                 = {coul_cutoff}\n")
            else:
                raise UnsupportedExportError(
                    f"Electrostatics method {self.coul_method} not supported",
                )

            if self.vdw_method == "cutoff":
                mdp.write("vdwtype                  = cutoff\n")
            elif self.vdw_method in ("Ewald3D", "pme", "PME", _PME):
                mdp.write("vdwtype                  = PME\n")
                # TODO: Wire this through like `ewald_tolerance` in `to_openmm`
                # TODO: Should this match electrostatics PME tolerance?
                mdp.write("ewald-rtol-lj            = 1e-4\n")
                mdp.write("lj-pme-comb-rule         = geometric\n")
            else:
                raise UnsupportedExportError(
                    f"vdW method {self.vdw_method} not supported",
                )

            vdw_cutoff = round(self.vdw_cutoff.m_as("nanometer"), 4)
            mdp.write(f"rvdw                     = {vdw_cutoff}\n")

            if self.switching_function and self.vdw_method == "cutoff":
                mdp.write("vdw-modifier             = Potential-switch\n")
                distance = round(self.switching_distance.m_as("nanometer"), 4)
                mdp.write(f"rvdw-switch              = {distance}\n")
            else:
                mdp.write("vdw-modifier             = None\n")
                mdp.write("rvdwswitch               = 0\n")

            # No errors, we can now write to disk!
            Path(mdp_file).write_text(mdp.getvalue())

    def write_lammps_input(
        self,
        interchange: "Interchange",
        input_file: str = "run.in",
        data_file: str = "out.lmp",
    ) -> None:
        """
        Write a LAMMPS input file for running single-point energies.

        Parameters
        ----------
        interchange
            The `Interchange` used when preparing this LAMMPS input.
        input_file
            The name of the file to write the input to, defaults to `run.in`.
        data_file
            The name of the file to write the data to, defaults to `out.lmp`.
        """
        # TODO: Get constrained angles
        # TODO: Process rigid water

        def _get_coeffs_of_constrained_bonds_and_angles(
            interchange: "Interchange",
        ) -> tuple[set[int], set[int]]:
            """
            Get coefficients of bonds and angles that appear to be constrained.

            Refactor this when LAMMPS export uses a dedicated class.

            * Coefficients are matched by stored SMIRKS
            * Coefficients are ints associated with Bond Coeffs / Angle Coeffs section
            * Coefficients are zero-indexed
            """
            constraint_styles = {key.associated_handler for key in interchange["Constraints"].potentials}

            if len(constraint_styles.difference({"Bonds", "Angles"})) > 0:
                raise NotImplementedError(
                    "Found unsupported constraints case in LAMMPS input writer.",
                )

            constrained_bond_smirks = {
                key.id for key in interchange["Constraints"].potentials if key.associated_handler == "Bonds"
            }

            constrained_angle_smirks = {
                key.id for key in interchange["Constraints"].potentials if key.associated_handler == "Angles"
            }

            return (
                {
                    key
                    for key, val in dict(
                        enumerate(interchange["Bonds"].potentials),
                    ).items()
                    if val.id in constrained_bond_smirks
                },
                {
                    key
                    for key, val in dict(
                        enumerate(interchange["Angles"].potentials),
                    ).items()
                    if val.id in constrained_angle_smirks
                },
            )

        # zero-indexed here
        (
            constrained_bond_coeffs,
            constrained_angle_coeffs,
        ) = _get_coeffs_of_constrained_bonds_and_angles(interchange)

        # Construct the input file in memory so nothing is written to disk if an
        # error is encountered.
        with StringIO() as lmp:
            if self.switching_function is not None:
                if self.switching_distance.m > 0.0:
                    warnings.warn(
                        f"A switching distance {self.switching_distance} was specified by the "
                        "force field, but LAMMPS may not implement a switching function as "
                        "specified by SMIRNOFF. Using a hard cut-off instead. Non-bonded "
                        "interactions will be affected.",
                        SwitchingFunctionNotImplementedWarning,
                    )

            lmp.write(
                "units real\natom_style full\n\ndimension 3\nboundary p p p\n\n",
            )

            if len(interchange["Bonds"].key_map) > 0:
                lmp.write("bond_style harmonic\n")

            if len(interchange["Angles"].key_map) > 0:
                lmp.write("angle_style harmonic\n")

            try:
                if len(interchange["ProperTorsions"].key_map) > 0:
                    lmp.write("dihedral_style fourier\n")
            except LookupError:
                # no torsions here
                pass

            try:
                if len(interchange["ImproperTorsions"].key_map) > 0:
                    lmp.write("improper_style cvff\n")
            except LookupError:
                # no impropers here
                pass

            # TODO: LAMMPS puts this information in the "run" file. Should it live in MDConfig or not?
            scale_factors = {
                "vdW": {
                    "1-2": 0.0,
                    "1-3": 0.0,
                    "1-4": 0.5,
                    "1-5": 1,
                },
                "Electrostatics": {
                    "1-2": 0.0,
                    "1-3": 0.0,
                    "1-4": 0.8333333333,
                    "1-5": 1,
                },
            }

            lmp.write(
                "special_bonds lj "
                f"{scale_factors['vdW']['1-2']} "
                f"{scale_factors['vdW']['1-3']} "
                f"{scale_factors['vdW']['1-4']} "
                "coul "
                f"{scale_factors['Electrostatics']['1-2']} "
                f"{scale_factors['Electrostatics']['1-3']} "
                f"{scale_factors['Electrostatics']['1-4']} "
                "\n",
            )

            vdw_cutoff = round(self.vdw_cutoff.m_as("angstrom"), 4)
            coul_cutoff = round(self.coul_cutoff.m_as("angstrom"), 4)

            if self.coul_method == _PME:
                lmp.write(f"pair_style lj/cut/coul/long {vdw_cutoff} {coul_cutoff}\n")
            elif self.coul_method == "cutoff":
                lmp.write(f"pair_style lj/cut/coul/cut {vdw_cutoff} {coul_cutoff}\n")
            else:
                raise UnsupportedExportError(
                    f"Unsupported electrostatics method {self.coul_method}",
                )

            if self.mixing_rule == "lorentz-berthelot":
                lmp.write("pair_modify mix arithmetic tail yes\n\n")
            elif self.mixing_rule == "geometric":
                lmp.write("pair_modify mix geometric tail yes\n\n")
            else:
                raise UnsupportedExportError(
                    f"Mixing rule {self.mixing_rule} not supported",
                )

            lmp.write(f"read_data {data_file}\n\n")
            lmp.write(
                "thermo_style custom ebond eangle edihed eimp epair evdwl ecoul elong etail pe\n\n",
            )

            if len(constrained_bond_coeffs.union(constrained_angle_coeffs)) > 0:
                # https://docs.lammps.org/fix_shake.html
                # TODO: Apply fix to just a group (sub-group)?
                lmp.write(
                    "fix 100 all shake 0.0001 20 10 ",
                )

                if constrained_bond_coeffs:
                    lmp.write(
                        f"b {' '.join([str(val + 1) for val in constrained_bond_coeffs])}",
                    )

                if constrained_angle_coeffs:
                    lmp.write(
                        f"a {' '.join([str(val + 1) for val in constrained_angle_coeffs])}",
                    )

                lmp.write("\n")

            if self.coul_method == _PME:
                # Note: LAMMPS will error out if using kspace on something with all zero charges,
                # so this may not work if all partial charges are zero
                lmp.write("kspace_style pppm 1e-4\n")

            lmp.write("run 0\n")

            # No errors, safe to write to disk!
            Path(input_file).write_text(lmp.getvalue())

    def write_sander_input_file(
        self,
        input_file: str = "run.in",
    ) -> None:
        """
        Write a Sander input file for running single-point energies.

        Parameters
        ----------
        input_file
            The name of the file to write the input to, defaults to `run.in`.

        """
        # Construct the file in memory so nothing is written to disk if an
        # error is encountered.
        with StringIO() as sander:
            sander.write("single-point energy\n&cntrl\nimin=1,\nmaxcyc=0,\nntb=1,\n")

            if self.switching_function is not None:
                if self.switching_distance.m > 0.0:
                    warnings.warn(
                        f"A switching distance {self.switching_distance} was specified by the "
                        "force field, but Amber does not implement a switching function. Using a "
                        "hard cut-off instead. Non-bonded interactions will be affected.",
                        SwitchingFunctionNotImplementedWarning,
                    )

                # Whether this is stored as zero or positive distance, pass a
                # negative value to ensure it's turned off.
                sander.write(f"fswitch={-1.0},\n")

            if self.constraints in ["none", None]:
                sander.write("ntc=1,\nntf=1,\n")
            # TODO: This is an approximation, but most of the time these will be set to 2
            #       Amber cannot ignore H-O-H angle energy without ignoring all H-X-X angles,
            #       See 21.7.1. in Amber22 manual
            elif self.constraints in ("h-bonds", "all-bonds", "all-angles"):
                sander.write(
                    "ntc=1,\n"  # do NOT perform shake, since it will modify positions, but ...
                    "ntf=2,\n",  # ... ignore interactions of bonds including hydrogen atoms
                )
            # TODO: Cover other cases, though hard to reach with mainline OpenFF force fields
            else:
                raise UnsupportedExportError(
                    f"Unclear how to apply {self.constraints} with sander",
                )

            if self.vdw_method == "cutoff":
                vdw_cutoff = round(self.vdw_cutoff.m_as("angstrom"), 4)
                sander.write(f"cut={vdw_cutoff},\n")
            else:
                raise NonperiodicNoCutoffNotSupportedError(
                    f"vdW method {self.vdw_method} not supported",
                )

            if self.coul_method == _PME:
                sander.write("/\n&ewald\norder=4\nskinnb=1.0\n/")

            sander.write("/\n")

            # No errors, safe to write to disk!
            Path(input_file).write_text(sander.getvalue())


def _infer_constraints(interchange: "Interchange") -> str:
    if "Constraints" not in interchange.collections:
        return "none"
    elif "Bonds" not in interchange.collections:
        return "none"
    else:
        num_constraints = len(interchange["Constraints"].key_map)
        if num_constraints == 0:
            return "none"
        else:
            num_h_bonds = _get_num_h_bonds(interchange.topology)

            num_bonds = len(interchange["Bonds"].key_map)
            num_angles = len(interchange["Angles"].key_map)

            if num_constraints == num_h_bonds:
                return "h-bonds"
            elif num_constraints == num_bonds:
                return "all-bonds"
            elif num_constraints == (num_bonds + num_angles):
                return "all-angles"

            else:
                # TODO: Rigid waters may not have bond and angle parameters, but still have 3 constraints
                #       per molecule. There should be a better way to process these, but it's non-trivial
                #       to detect water molecules in a performance, scalable way without false positives.
                warnings.warn(
                    "Ambiguous failure while processing constraints. Constraining h-bonds as a stopgap.",
                )

                return "h-bonds"


def get_smirnoff_defaults(periodic: bool = False) -> MDConfig:
    """
    Return an `MDConfig` object that matches settings used in SMIRNOFF force fields (through Sage).

    Parameters
    ----------
    periodic
        Whether to use periodic boundary conditions.

    Returns
    -------
    config
        An `MDConfig` object with settings that match those used in SMIRNOFF force fields.

    """
    return MDConfig(
        periodic=periodic,
        constraints="h-bonds",
        vdw_method="cutoff",
        vdw_cutoff=Quantity(0.9, "nanometer"),
        mixing_rule="lorentz-berthelot",
        switching_function=True,
        switching_distance=Quantity(0.8, "nanometer"),
        coul_method="PME" if periodic else "Coulomb",
    )


def get_intermol_defaults(periodic: bool = False) -> MDConfig:
    """
    Return an `MDConfig` object that attempts to match settings used in InterMol tests.

    These settings are poor choices for production but can be useful for testing. See also
        - 10.1007/s10822-016-9977-1
        - https://github.com/shirtsgroup/InterMol/blob/master/intermol/tests/
            /gromacs/grompp_vacuum.mdp
            /lammps/unit_tests/atom_style-full_vacuum/atom_style-full-data_vacuum.input
            /amber/min_vacuum.in

    Parameters
    ----------
    periodic
        Whether to use periodic boundary conditions.

    Returns
    -------
    config
        An `MDConfig` object with settings that match those used in InterMol tests.

    """
    return MDConfig(
        periodic=periodic,
        constraints="none",
        vdw_method="cutoff",
        vdw_cutoff=Quantity(0.9, "nanometer"),
        mixing_rule="lorentz-berthelot",
        switching_function=False,
        switching_distance=Quantity(0.0, "angstrom"),
        coul_method="PME" if periodic else "cutoff",
        coul_cutoff=Quantity(0.9, "nanometer") if periodic else Quantity(2.0, "nanometer"),
    )
